# stdlib
from collections.abc import Callable
from enum import Enum
import hashlib
import inspect
import logging
from typing import Any

# third party
from pydantic import model_validator
from result import Err
from result import Ok
from result import Result
from typing_extensions import Self

# relative
from ...abstract_server import ServerSideType
from ...client.api import APIRegistry
from ...client.client import SyftClient
from ...custom_worker.config import DockerWorkerConfig
from ...custom_worker.config import WorkerConfig
from ...custom_worker.k8s import IN_KUBERNETES
from ...serde.serializable import serializable
from ...serde.serialize import _serialize
from ...server.credentials import SyftVerifyKey
from ...store.linked_obj import LinkedObject
from ...types.datetime import DateTime
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.syncable_object import SyncableSyftObject
from ...types.transforms import TransformContext
from ...types.transforms import add_server_uid_for_key
from ...types.transforms import generate_id
from ...types.transforms import transform
from ...types.twin_object import TwinObject
from ...types.uid import LineageID
from ...types.uid import UID
from ...util import options
from ...util.colors import SURFACE
from ...util.decorators import deprecated
from ...util.markdown import markdown_as_class_with_fields
from ...util.notebook_ui.icons import Icon
from ...util.util import prompt_warning_message
from ..action.action_object import ActionObject
from ..action.action_service import ActionService
from ..action.action_store import ActionObjectPermission
from ..action.action_store import ActionPermission
from ..blob_storage.service import BlobStorageService
from ..code.user_code import UserCode
from ..code.user_code import UserCodeStatus
from ..code.user_code import UserCodeStatusCollection
from ..context import AuthedServiceContext
from ..context import ChangeContext
from ..job.job_stash import Job
from ..job.job_stash import JobStatus
from ..notification.notifications import Notification
from ..response import SyftError
from ..response import SyftSuccess
from ..user.user import UserView

logger = logging.getLogger(__name__)


@serializable(canonical_name="RequestStatus", version=1)
class RequestStatus(Enum):
    PENDING = 0
    REJECTED = 1
    APPROVED = 2

    @classmethod
    def from_usercode_status(cls, status: UserCodeStatusCollection) -> "RequestStatus":
        if status.approved:
            return RequestStatus.APPROVED
        elif status.denied:
            return RequestStatus.REJECTED
        else:
            return RequestStatus.PENDING


@serializable()
class Change(SyftObject):
    __canonical_name__ = "Change"
    __version__ = SYFT_OBJECT_VERSION_1

    linked_obj: LinkedObject | None = None

    def change_object_is_type(self, type_: type) -> bool:
        return self.linked_obj is not None and type_ == self.linked_obj.object_type


@serializable()
class ChangeStatus(SyftObject):
    __canonical_name__ = "ChangeStatus"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID | None = None  # type: ignore[assignment]
    change_id: UID
    applied: bool = False

    @classmethod
    def from_change(cls, change: Change, applied: bool) -> Self:
        return cls(change_id=change.id, applied=applied)


@serializable()
class ActionStoreChange(Change):
    __canonical_name__ = "ActionStoreChange"
    __version__ = SYFT_OBJECT_VERSION_1

    linked_obj: LinkedObject
    apply_permission_type: ActionPermission

    __repr_attrs__ = ["linked_obj", "apply_permission_type"]

    def _run(
        self, context: ChangeContext, apply: bool
    ) -> Result[SyftSuccess, SyftError]:
        try:
            action_service: ActionService = context.server.get_service(ActionService)  # type: ignore[assignment]
            blob_storage_service = context.server.get_service(BlobStorageService)
            action_store = action_service.store

            # can we ever have a lineage ID in the store?
            obj_uid = self.linked_obj.object_uid
            obj_uid = obj_uid.id if isinstance(obj_uid, LineageID) else obj_uid

            action_obj = action_store.get(
                uid=obj_uid,
                credentials=context.approving_user_credentials,
            )

            if action_obj.is_err():
                return Err(SyftError(message=f"{action_obj.err()}"))

            action_obj = action_obj.ok()

            owner_permission = ActionObjectPermission(
                uid=obj_uid,
                credentials=context.approving_user_credentials,
                permission=self.apply_permission_type,
            )
            if action_store.has_permission(permission=owner_permission):
                id_action = (
                    action_obj.id
                    if not isinstance(action_obj.id, LineageID)
                    else action_obj.id.id
                )
                requesting_permission_action_obj = ActionObjectPermission(
                    uid=id_action,
                    credentials=context.requesting_user_credentials,
                    permission=self.apply_permission_type,
                )
                if isinstance(action_obj, TwinObject):
                    uid_blob = action_obj.private.syft_blob_storage_entry_id
                else:
                    uid_blob = action_obj.syft_blob_storage_entry_id
                requesting_permission_blob_obj = (
                    ActionObjectPermission(
                        uid=uid_blob,
                        credentials=context.requesting_user_credentials,
                        permission=self.apply_permission_type,
                    )
                    if uid_blob
                    else None
                )
                if apply:
                    logger.debug(
                        "ADDING PERMISSION", requesting_permission_action_obj, id_action
                    )
                    action_store.add_permission(requesting_permission_action_obj)
                    (
                        blob_storage_service.stash.add_permission(
                            requesting_permission_blob_obj
                        )
                        if requesting_permission_blob_obj
                        else None
                    )
                else:
                    if action_store.has_permission(requesting_permission_action_obj):
                        action_store.remove_permission(requesting_permission_action_obj)
                    if (
                        requesting_permission_blob_obj
                        and blob_storage_service.stash.has_permission(
                            requesting_permission_blob_obj
                        )
                    ):
                        blob_storage_service.stash.remove_permission(
                            requesting_permission_blob_obj
                        )
            else:
                return Err(
                    SyftError(
                        message=f"No permission for approving_user_credentials {context.approving_user_credentials}"
                    )
                )
            return Ok(SyftSuccess(message=f"{type(self)} Success"))
        except Exception as e:
            logger.error(f"failed to apply {type(self)}", exc_info=e)
            return Err(SyftError(message=str(e)))

    def apply(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=True)

    def undo(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=False)

    def __repr_syft_nested__(self) -> str:
        return f"Apply <b>{self.apply_permission_type}</b> to \
<i>{self.linked_obj.object_type.__canonical_name__}:{self.linked_obj.object_uid.short()}</i>."


@serializable()
class CreateCustomImageChange(Change):
    __canonical_name__ = "CreateCustomImageChange"
    __version__ = SYFT_OBJECT_VERSION_1

    config: WorkerConfig
    tag: str | None = None
    registry_uid: UID | None = None
    pull_image: bool = True

    __repr_attrs__ = ["config", "tag"]

    @model_validator(mode="after")
    def _tag_required_for_dockerworkerconfig(self) -> Self:
        if isinstance(self.config, DockerWorkerConfig) and self.tag is None:
            raise ValueError("`tag` is required for `DockerWorkerConfig`.")
        return self

    def _run(
        self, context: ChangeContext, apply: bool
    ) -> Result[SyftSuccess, SyftError]:
        try:
            worker_image_service = context.server.get_service("SyftWorkerImageService")

            service_context = context.to_service_ctx()
            result = worker_image_service.submit(
                service_context, worker_config=self.config
            )

            if isinstance(result, SyftError):
                return Err(result)

            result = worker_image_service.stash.get_by_worker_config(
                service_context.credentials, config=self.config
            )

            if result.is_err():
                return Err(SyftError(message=f"{result.err()}"))

            if (worker_image := result.ok()) is None:
                return Err(SyftError(message="The worker image does not exist."))

            build_success_message = "Image was pre-built."

            if not worker_image.is_prebuilt:
                build_result = worker_image_service.build(
                    service_context,
                    image_uid=worker_image.id,
                    tag=self.tag,
                    registry_uid=self.registry_uid,
                    pull_image=self.pull_image,
                )

                if isinstance(build_result, SyftError):
                    return Err(build_result)

                build_success_message = build_result.message

            build_success = SyftSuccess(
                message=f"Build result: {build_success_message}"
            )

            if IN_KUBERNETES and not worker_image.is_prebuilt:
                push_result = worker_image_service.push(
                    service_context,
                    image_uid=worker_image.id,
                    username=context.extra_kwargs.get("registry_username", None),
                    password=context.extra_kwargs.get("registry_password", None),
                )

                if isinstance(push_result, SyftError):
                    return Err(push_result)

                return Ok(
                    SyftSuccess(
                        message=f"{build_success}\nPush result: {push_result.message}"
                    )
                )

            return Ok(build_success)

        except Exception as e:
            return Err(SyftError(message=f"Failed to create/build image: {e}"))

    def apply(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=True)

    def undo(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=False)

    def __repr_syft_nested__(self) -> str:
        return f"Create Image for Config: {self.config} with tag: {self.tag}"


@serializable()
class CreateCustomWorkerPoolChange(Change):
    __canonical_name__ = "CreateCustomWorkerPoolChange"
    __version__ = SYFT_OBJECT_VERSION_1

    pool_name: str
    num_workers: int
    image_uid: UID | None = None
    config: WorkerConfig | None = None
    pod_annotations: dict[str, str] | None = None
    pod_labels: dict[str, str] | None = None

    __repr_attrs__ = ["pool_name", "num_workers", "image_uid"]

    def _run(
        self, context: ChangeContext, apply: bool
    ) -> Result[SyftSuccess, SyftError]:
        """
        This function is run when the DO approves (apply=True)
        or deny (apply=False) the request.
        """
        # TODO: refactor the returned Err(SyftError) or Ok(SyftSuccess) to just
        # SyftError or SyftSuccess
        if apply:
            # get the worker pool service and try to launch a pool
            worker_pool_service = context.server.get_service("SyftWorkerPoolService")
            service_context: AuthedServiceContext = context.to_service_ctx()

            if self.config is not None:
                result = worker_pool_service.image_stash.get_by_worker_config(
                    service_context.credentials, self.config
                )
                if result.is_err():
                    return Err(SyftError(message=f"{result.err()}"))
                worker_image = result.ok()
                self.image_uid = worker_image.id

            result = worker_pool_service.launch(
                context=service_context,
                pool_name=self.pool_name,
                image_uid=self.image_uid,
                num_workers=self.num_workers,
                registry_username=context.extra_kwargs.get("registry_username", None),
                registry_password=context.extra_kwargs.get("registry_password", None),
                pod_annotations=self.pod_annotations,
                pod_labels=self.pod_labels,
            )
            if isinstance(result, SyftError):
                return Err(result)
            else:
                return Ok(result)
        else:
            return Err(
                SyftError(
                    message=f"Request to create a worker pool with name {self.name} denied"
                )
            )

    def apply(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=True)

    def undo(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=False)

    def __repr_syft_nested__(self) -> str:
        return (
            f"Create Worker Pool '{self.pool_name}' for Image with id {self.image_uid}"
        )


@serializable()
class Request(SyncableSyftObject):
    __canonical_name__ = "Request"
    __version__ = SYFT_OBJECT_VERSION_1

    requesting_user_verify_key: SyftVerifyKey
    requesting_user_name: str = ""
    requesting_user_email: str | None = ""
    requesting_user_institution: str | None = ""
    approving_user_verify_key: SyftVerifyKey | None = None
    request_time: DateTime
    updated_at: DateTime | None = None
    server_uid: UID
    request_hash: str
    changes: list[Change]
    history: list[ChangeStatus] = []
    tags: list[str] = []

    __table_coll_widths__ = [
        "min-content",
        "auto",
        "auto",
        "auto",
        "auto",
        "auto",
    ]

    __attr_searchable__ = [
        "requesting_user_verify_key",
        "approving_user_verify_key",
    ]
    __attr_unique__ = ["request_hash"]
    __repr_attrs__ = [
        "request_time",
        "updated_at",
        "status",
        "changes",
        "requesting_user_verify_key",
    ]
    __exclude_sync_diff_attrs__ = ["server_uid", "changes", "history"]
    __table_sort_attr__ = "Request time"

    def _repr_html_(self) -> Any:
        # add changes
        updated_at_line = ""
        if self.updated_at is not None:
            updated_at_line += (
                f"<p><strong>Created by: </strong>{self.requesting_user_name}</p>"
            )
        str_changes_ = []
        for change in self.changes:
            str_change = (
                change.__repr_syft_nested__()
                if hasattr(change, "__repr_syft_nested__")
                else type(change)
            )
            str_change = f"{str_change}. "
            str_changes_.append(str_change)
        str_changes = "\n".join(str_changes_)
        api = APIRegistry.api_for(
            self.server_uid,
            self.syft_client_verify_key,
        )
        shared_with_line = ""
        if self.code and len(self.code.output_readers) > 0:
            # owner_names = ["canada", "US"]
            owners_string = " and ".join(
                [f"<strong>{x}</strong>" for x in self.code.output_reader_names]  # type: ignore
            )
            shared_with_line += (
                f"<p><strong>Custom Policy: </strong> "
                f"outputs are <strong>shared</strong> with the owners of {owners_string} once computed"
            )

        server_info = ""
        if api is not None:
            metadata = api.services.metadata.get_metadata()
            server_name = (
                api.server_name.capitalize() if api.server_name is not None else ""
            )
            server_type = metadata.server_type.value.capitalize()
            server_info = f"<p><strong>Requested on: </strong> {server_name} of type <strong>{server_type}</strong></p>"

        email_str = (
            f"({self.requesting_user_email})" if self.requesting_user_email else ""
        )
        institution_str = (
            f"<strong>Institution:</strong> {self.requesting_user_institution}"
            if self.requesting_user_institution
            else ""
        )

        return f"""
            <style>
            .syft-request {{color: {SURFACE[options.color_theme]};}}
            </style>
            <div class='syft-request'>
                <h3>Request</h3>
                <p><strong>Id: </strong>{self.id}</p>
                <p><strong>Request time: </strong>{self.request_time}</p>
                {updated_at_line}
                {shared_with_line}
                <p><strong>Status: </strong>{self.status}</p>
                {server_info}
                <p><strong>Requested by:</strong> {self.requesting_user_name} {email_str} {institution_str}</p>
                <p><strong>Changes: </strong> {str_changes}</p>
            </div>

            """

    @property
    def html_description(self) -> str:
        desc = " ".join([x.__repr_syft_nested__() for x in self.changes])
        # desc = desc.replace('\n', '')
        # desc = desc.replace('<br>', '\n')
        desc = desc.replace(". ", ".\n\n")
        desc = desc.replace("<b>", "")
        desc = desc.replace("</b>", "")
        desc = desc.replace("<i>", "")
        desc = desc.replace("</i>", "")

        return desc

    def _coll_repr_(self) -> dict[str, str | dict[str, str]]:
        if self.status == RequestStatus.APPROVED:
            badge_color = "badge-green"
        elif self.status == RequestStatus.PENDING:
            badge_color = "badge-gray"
        else:
            badge_color = "badge-red"

        status_badge = {"value": self.status.name.capitalize(), "type": badge_color}

        user_data = [
            self.requesting_user_name,
            self.requesting_user_email,
            self.requesting_user_institution,
        ]

        return {
            "Description": self.html_description,
            "Requested By": "\n".join(user_data),
            "Creation Time": str(self.request_time),
            "Status": status_badge,
        }

    @property
    def code_id(self) -> UID:
        for change in self.changes:
            if isinstance(change, UserCodeStatusChange):
                return change.linked_user_code.object_uid
        return SyftError(
            message="This type of request does not have code associated with it."
        )

    @property
    def codes(self) -> Any:
        for change in self.changes:
            if isinstance(change, UserCodeStatusChange):
                return change.codes
        return SyftError(
            message="This type of request does not have code associated with it."
        )

    def get_user_code(self, context: AuthedServiceContext) -> UserCode | None:
        for change in self.changes:
            if isinstance(change, UserCodeStatusChange):
                return change.get_user_code(context)
        return None

    @property
    def code(self) -> UserCode | SyftError:
        for change in self.changes:
            if isinstance(change, UserCodeStatusChange):
                return change.code
        return SyftError(
            message="This type of request does not have code associated with it."
        )

    @property
    def current_change_state(self) -> dict[UID, bool]:
        change_applied_map = {}
        for change_status in self.history:
            # only store the last change
            change_applied_map[change_status.change_id] = change_status.applied

        return change_applied_map

    @property
    def icon(self) -> str:
        return Icon.REQUEST.svg

    def get_status(self, context: AuthedServiceContext | None = None) -> RequestStatus:
        is_l0_deployment = (
            self.get_is_l0_deployment(context) if context else self.is_l0_deployment
        )
        if is_l0_deployment:
            code_status = self.code.get_status(context) if context else self.code.status
            return RequestStatus.from_usercode_status(code_status)

        if len(self.history) == 0:
            return RequestStatus.PENDING

        all_changes_applied = all(self.current_change_state.values()) and (
            len(self.current_change_state) == len(self.changes)
        )

        request_status = (
            RequestStatus.APPROVED if all_changes_applied else RequestStatus.REJECTED
        )

        return request_status

    @property
    def status(self) -> RequestStatus:
        return self.get_status()

    def approve(
        self,
        disable_warnings: bool = False,
        approve_nested: bool = False,
        **kwargs: dict,
    ) -> Result[SyftSuccess, SyftError]:
        api = self._get_api()
        if isinstance(api, SyftError):
            return api

        if self.is_l0_deployment:
            return SyftError(
                message="This request is a low-side request. Please sync your results to approve."
            )
        # TODO: Refactor so that object can also be passed to generate warnings
        if api.connection:
            metadata = api.connection.get_server_metadata(api.signing_key)
        else:
            metadata = None
        message = None

        is_code_request = not isinstance(self.codes, SyftError)

        if is_code_request and len(self.codes) > 1 and not approve_nested:
            return SyftError(
                message="Multiple codes detected, please use approve_nested=True"
            )

        if metadata and metadata.server_side_type == ServerSideType.HIGH_SIDE.value:
            message = (
                "You're approving a request on "
                f"{metadata.server_side_type} side {metadata.server_type} "
                "which may host datasets with private information."
            )
        if message and metadata and metadata.show_warnings and not disable_warnings:
            prompt_warning_message(message=message, confirm=True)
        msg = (
            "Approving request ",
            f"on change {self.code.service_func_name} " if is_code_request else "",
            f"for datasite {api.server_name}",
        )

        print("".join(msg))
        res = api.services.request.apply(self.id, **kwargs)
        return res

    def deny(self, reason: str) -> SyftSuccess | SyftError:
        """Denies the particular request.

        Args:
            reason (str): Reason for which the request has been denied.
        """
        api = self._get_api()
        if isinstance(api, SyftError):
            return api

        if self.is_l0_deployment:
            if self.status == RequestStatus.APPROVED:
                prompt_warning_message(
                    "This request already has results published to the data scientist. "
                    "They will still be able to access those results."
                )
            result = api.code.update(id=self.code_id, l0_deny_reason=reason)
            if isinstance(result, SyftError):
                return result
            return SyftSuccess(message=f"Request denied with reason: {reason}")

        return api.services.request.undo(uid=self.id, reason=reason)

    @property
    def is_l0_deployment(self) -> bool:
        return bool(self.code) and self.code.is_l0_deployment

    def get_is_l0_deployment(self, context: AuthedServiceContext) -> bool:
        code = self.get_user_code(context)
        if code:
            return code.is_l0_deployment
        else:
            return False

    def approve_with_client(self, client: SyftClient) -> Result[SyftSuccess, SyftError]:
        if self.is_l0_deployment:
            return SyftError(
                message="This request is a low-side request. Please sync your results to approve."
            )

        print(f"Approving request for datasite {client.name}")
        return client.api.services.request.apply(self.id)

    def apply(self, context: AuthedServiceContext) -> Result[SyftSuccess, SyftError]:
        change_context: ChangeContext = ChangeContext.from_service(context)
        change_context.requesting_user_credentials = self.requesting_user_verify_key
        for change in self.changes:
            # by default change status is not applied
            change_status = ChangeStatus(change_id=change.id, applied=False)
            result = change.apply(context=change_context)
            if isinstance(result, SyftError):
                return result
            if result.is_err():
                # add to history and save history to request
                self.history.append(change_status)
                self.save(context=context)
                return result

            # If no error, then change successfully applied.
            change_status.applied = True
            self.history.append(change_status)

        self.updated_at = DateTime.now()
        self.save(context=context)

        return Ok(SyftSuccess(message=f"Request {self.id} changes applied"))

    def undo(self, context: AuthedServiceContext) -> Result[SyftSuccess, SyftError]:
        change_context: ChangeContext = ChangeContext.from_service(context)
        change_context.requesting_user_credentials = self.requesting_user_verify_key

        current_change_state = self.current_change_state
        for change in self.changes:
            # by default change status is not applied
            is_change_applied = current_change_state.get(change.id, False)
            change_status = ChangeStatus(
                change_id=change.id,
                applied=is_change_applied,
            )
            # undo here may be deny for certain Changes (UserCodeChange)
            result = change.undo(context=change_context)
            if result.is_err():
                # add to history and save history to request
                self.history.append(change_status)
                self.save(context=context)
                return result

            # If no error, then change successfully undone.
            change_status.applied = False
            self.history.append(change_status)

        self.updated_at = DateTime.now()
        result = self.save(context=context)
        if isinstance(result, SyftError):
            return Err(result)
        # override object with latest changes.
        self = result
        return Ok(SyftSuccess(message=f"Request {self.id} changes undone."))

    def save(self, context: AuthedServiceContext) -> Result[SyftSuccess, SyftError]:
        # relative
        from .request_service import RequestService

        save_method = context.server.get_service_method(RequestService.save)
        return save_method(context=context, request=self)

    def _create_action_object_for_deposited_result(
        self,
        result: Any,
    ) -> ActionObject | SyftError:
        api = self._get_api()
        if isinstance(api, SyftError):
            return api

        # Ensure result is an ActionObject
        if isinstance(result, ActionObject):
            existing_job = api.services.job.get_by_result_id(result.id.id)
            if existing_job is not None:
                return SyftError(
                    message=f"This ActionObject is already the result of Job {existing_job.id}"
                )
            action_object = result
        else:
            action_object = ActionObject.from_obj(
                result,
                syft_client_verify_key=self.syft_client_verify_key,
                syft_server_location=self.syft_server_location,
            )

        # Ensure ActionObject exists on this server
        action_object_is_from_this_server = isinstance(
            api.services.action.exists(action_object.id.id), SyftSuccess
        )
        if (
            action_object.syft_blob_storage_entry_id is None
            or not action_object_is_from_this_server
        ):
            action_object.reload_cache()
            result = action_object._send(self.server_uid, self.syft_client_verify_key)
            if isinstance(result, SyftError):
                return result

        return action_object

    def _create_output_history_for_deposited_result(
        self, job: Job, result: Any
    ) -> SyftSuccess | SyftError:
        code = self.code
        if isinstance(code, SyftError):
            return code
        api = self._get_api()
        if isinstance(api, SyftError):
            return api

        input_ids = {}
        input_policy = code.input_policy
        if input_policy is not None:
            for input_ in input_policy.inputs.values():
                input_ids.update(input_)
        res = api.services.code.store_execution_output(
            user_code_id=code.id,
            outputs=result,
            job_id=job.id,
            input_ids=input_ids,
        )

        return res

    def deposit_result(
        self,
        result: Any,
        log_stdout: str = "",
        log_stderr: str = "",
    ) -> Job | SyftError:
        """
        Adds a result to this Request:
        - Create an ActionObject from the result (if not already an ActionObject)
        - Ensure ActionObject exists on this server
        - Create Job with new result and logs
        - Update the output history

        Args:
            result (Any): ActionObject or any object to be saved as an ActionObject.
            logs (str | None, optional): Optional logs to be saved with the Job. Defaults to None.

        Returns:
            Job | SyftError: Job object if successful, else SyftError.
        """

        # TODO check if this is a low-side request. If not, SyftError

        api = self._get_api()
        if isinstance(api, SyftError):
            return api
        code = self.code
        if isinstance(code, SyftError):
            return code

        if not self.is_l0_deployment:
            return SyftError(
                message="deposit_result is only available for low side code requests. "
                "Please use request.approve() instead."
            )

        # Create ActionObject
        action_object = self._create_action_object_for_deposited_result(result)
        if isinstance(action_object, SyftError):
            return action_object

        # Create Job
        # NOTE code owner read permissions are added when syncing this Job
        job = api.services.job.create_job_for_user_code_id(
            code.id,
            result=action_object,
            log_stdout=log_stdout,
            log_stderr=log_stderr,
            status=JobStatus.COMPLETED,
            add_code_owner_read_permissions=False,
        )
        if isinstance(job, SyftError):
            return job

        # Add to output history
        res = self._create_output_history_for_deposited_result(job, result)
        if isinstance(res, SyftError):
            return res

        return job

    @deprecated(
        return_syfterror=True,
        reason="accept_by_depositing_result has been removed. Use approve instead to "
        "approve this request, or deposit_result to deposit a new result.",
    )
    def accept_by_depositing_result(self, result: Any, force: bool = False) -> Any:
        pass

    def get_sync_dependencies(
        self, context: AuthedServiceContext
    ) -> list[UID] | SyftError:
        dependencies = []
        code_id = self.code_id
        if isinstance(code_id, SyftError):
            return code_id
        dependencies.append(code_id)

        return dependencies


@serializable()
class RequestInfo(SyftObject):
    # version
    __canonical_name__ = "RequestInfo"
    __version__ = SYFT_OBJECT_VERSION_1

    user: UserView
    request: Request
    message: Notification


@serializable()
class RequestInfoFilter(SyftObject):
    # version
    __canonical_name__ = "RequestInfoFilter"
    __version__ = SYFT_OBJECT_VERSION_1

    name: str | None = None


@serializable()
class SubmitRequest(SyftObject):
    __canonical_name__ = "SubmitRequest"
    __version__ = SYFT_OBJECT_VERSION_1

    changes: list[Change]
    requesting_user_verify_key: SyftVerifyKey | None = None


def hash_changes(context: TransformContext) -> TransformContext:
    if context.output is None:
        return context

    request_time = context.output["request_time"]
    key = context.output["requesting_user_verify_key"]
    changes = context.output["changes"]

    time_hash = hashlib.sha256(
        _serialize(request_time.utc_timestamp, to_bytes=True)
    ).digest()
    key_hash = hashlib.sha256(bytes(key.verify_key)).digest()
    changes_hash = hashlib.sha256(_serialize(changes, to_bytes=True)).digest()
    final_hash = hashlib.sha256(time_hash + key_hash + changes_hash).hexdigest()

    context.output["request_hash"] = final_hash

    return context


def add_request_time(context: TransformContext) -> TransformContext:
    if context.output is not None:
        context.output["request_time"] = DateTime.now()
    return context


def check_requesting_user_verify_key(context: TransformContext) -> TransformContext:
    if context.output and context.server and context.obj:
        if context.obj.requesting_user_verify_key and context.server.is_root(
            context.credentials
        ):
            context.output["requesting_user_verify_key"] = (
                context.obj.requesting_user_verify_key
            )
        else:
            context.output["requesting_user_verify_key"] = context.credentials

    return context


def add_requesting_user_info(context: TransformContext) -> TransformContext:
    if context.output is not None and context.server is not None:
        try:
            user_key = context.output["requesting_user_verify_key"]
            user_service = context.server.get_service("UserService")
            user = user_service.get_by_verify_key(user_key)
            context.output["requesting_user_name"] = user.name
            context.output["requesting_user_email"] = user.email
            context.output["requesting_user_institution"] = (
                user.institution if user.institution else ""
            )
        except Exception:
            context.output["requesting_user_name"] = "guest_user"

    return context


@transform(SubmitRequest, Request)
def submit_request_to_request() -> list[Callable]:
    return [
        generate_id,
        add_server_uid_for_key("server_uid"),
        add_request_time,
        check_requesting_user_verify_key,
        add_requesting_user_info,
        hash_changes,
    ]


@serializable()
class ObjectMutation(Change):
    __canonical_name__ = "ObjectMutation"
    __version__ = SYFT_OBJECT_VERSION_1

    linked_obj: LinkedObject | None = None
    attr_name: str
    value: Any | None = None
    match_type: bool
    previous_value: Any | None = None

    __repr_attrs__ = ["linked_obj", "attr_name"]

    def mutate(self, obj: Any, value: Any | None = None) -> Any:
        # check if attribute is a property setter first
        # this seems necessary for pydantic types
        attr = getattr(type(obj), self.attr_name, None)
        if inspect.isdatadescriptor(attr):
            if hasattr(attr, "fget") and hasattr(attr, "fset"):
                self.previous_value = attr.fget(obj)
                attr.fset(obj, value)
        else:
            self.previous_value = getattr(obj, self.attr_name, None)
            setattr(obj, self.attr_name, value)
        return obj

    def __repr_syft_nested__(self) -> str:
        return f"Mutate <b>{self.attr_name}</b> to <b>{self.value}</b>"

    def _run(
        self, context: ChangeContext, apply: bool
    ) -> Result[SyftSuccess, SyftError]:
        if self.linked_obj is None:
            return Err(SyftError(message=f"{self}'s linked object is None"))
        try:
            obj = self.linked_obj.resolve_with_context(context)
            if obj.is_err():
                return Err(SyftError(message=obj.err()))
            obj = obj.ok()
            if apply:
                obj = self.mutate(obj, value=self.value)
                self.linked_obj.update_with_context(context, obj)
            else:
                # unset the set value
                obj = self.mutate(obj, value=self.previous_value)
                self.linked_obj.update_with_context(context, obj)

            return Ok(SyftSuccess(message=f"{type(self)} Success"))
        except Exception as e:
            print(f"failed to apply {type(self)}. {e}")
            return Err(SyftError(message=str(e)))

    def apply(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=True)

    def undo(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=False)


def type_for_field(object_type: type, attr_name: str) -> type | None:
    field_type = None
    try:
        field_type = object_type.__dict__["__annotations__"][attr_name]
    except Exception:  # nosec
        try:
            field_type = object_type.__fields__.get(attr_name, None).type_
        except Exception:  # nosec
            pass
    return field_type


@serializable()
class EnumMutation(ObjectMutation):
    __canonical_name__ = "EnumMutation"
    __version__ = SYFT_OBJECT_VERSION_1

    enum_type: type[Enum]
    value: Enum | None = None
    match_type: bool = True

    __repr_attrs__ = ["linked_obj", "attr_name", "value"]

    @property
    def valid(self) -> SyftSuccess | SyftError:
        if self.match_type and not isinstance(self.value, self.enum_type):
            return SyftError(
                message=f"{type(self.value)} must be of type: {self.enum_type}"
            )
        return SyftSuccess(message=f"{type(self)} valid")

    @staticmethod
    def from_obj(
        linked_obj: LinkedObject, attr_name: str, value: Enum | None = None
    ) -> "EnumMutation":
        enum_type = type_for_field(linked_obj.object_type, attr_name)
        return EnumMutation(
            linked_obj=linked_obj,
            attr_name=attr_name,
            enum_type=enum_type,
            value=value,
            match_type=True,
        )

    def _run(
        self, context: ChangeContext, apply: bool
    ) -> Result[SyftSuccess, SyftError]:
        try:
            valid = self.valid
            if not valid:
                return Err(valid)
            if self.linked_obj is None:
                return Err(SyftError(message=f"{self}'s linked object is None"))
            obj = self.linked_obj.resolve_with_context(context)
            if obj.is_err():
                return Err(SyftError(message=obj.err()))
            obj = obj.ok()
            if apply:
                obj = self.mutate(obj=obj)

                self.linked_obj.update_with_context(context, obj)
            else:
                raise NotImplementedError
            return Ok(SyftSuccess(message=f"{type(self)} Success"))
        except Exception as e:
            print(f"failed to apply {type(self)}. {e}")
            return Err(SyftError(message=e))

    def apply(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=True)

    def undo(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=False)

    def __repr_syft_nested__(self) -> str:
        return f"Mutate <b>{self.enum_type}</b> to <b>{self.value}</b>"

    @property
    def link(self) -> SyftObject | None:
        if self.linked_obj:
            return self.linked_obj.resolve
        return None


@serializable()
class UserCodeStatusChange(Change):
    __canonical_name__ = "UserCodeStatusChange"
    __version__ = SYFT_OBJECT_VERSION_1

    value: UserCodeStatus
    linked_obj: LinkedObject
    linked_user_code: LinkedObject
    nested_solved: bool = False
    match_type: bool = True
    __repr_attrs__ = [
        "code.service_func_name",
        "code.input_policy_type.__canonical_name__",
        "code.output_policy_type.__canonical_name__",
        "code.worker_pool_name",
        "code.status.approved",
    ]

    @property
    def code(self) -> UserCode:
        if self.linked_user_code._resolve_cache:
            return self.linked_user_code._resolve_cache
        return self.linked_user_code.resolve

    def get_user_code(self, context: AuthedServiceContext) -> UserCode:
        resolve = self.linked_user_code.resolve_with_context(context)
        return resolve.ok()

    @property
    def codes(self) -> list[UserCode]:
        def recursive_code(server: Any) -> list:
            codes = []
            for obj, new_server in server.values():
                codes.append(obj.resolve)
                codes.extend(recursive_code(new_server))
            return codes

        codes = [self.code]
        codes.extend(recursive_code(self.code.nested_codes))
        return codes

    def nested_repr(self, server: Any | None = None, level: int = 0) -> str:
        msg = ""
        if server is None:
            server = self.code.nested_codes

        for service_func_name, (_, new_server) in server.items():  # type: ignore
            msg = "├──" + "──" * level + f"{service_func_name}<br>"
            msg += self.nested_repr(server=new_server, level=level + 1)
        return msg

    def __repr_syft_nested__(self) -> str:
        msg = (
            f"Request to change <strong>{self.code.service_func_name}</strong> "
            f"(Pool Id: <strong>{self.code.worker_pool_name}</strong>) "
        )
        msg += "to permission RequestStatus.APPROVED."
        if self.code.nested_codes is None or self.code.nested_codes == {}:  # type: ignore
            msg += " No nested requests"
        else:
            if self.nested_solved:
                # else:
                msg += "<br><br>This change requests the following nested functions calls:<br>"
                msg += self.nested_repr()
            else:
                msg += " Nested Requests not resolved"
        return msg

    def _repr_markdown_(self, wrap_as_python: bool = True, indent: int = 0) -> str:
        code = self.code
        input_policy_type = (
            code.input_policy_type.__canonical_name__
            if code.input_policy_type is not None
            else None
        )
        output_policy_type = (
            code.output_policy_type.__canonical_name__
            if code.output_policy_type is not None
            else None
        )
        repr_dict = {
            "function": code.service_func_name,
            "input_policy_type": f"{input_policy_type}",
            "output_policy_type": f"{output_policy_type}",
            "approved": f"{code.status.approved}",
        }
        return markdown_as_class_with_fields(self, repr_dict)

    @property
    def approved(self) -> bool:
        return self.linked_obj.resolve.approved

    @property
    def valid(self) -> SyftSuccess | SyftError:
        if self.match_type and not isinstance(self.value, UserCodeStatus):
            # TODO: fix the mypy issue
            return SyftError(  # type: ignore[unreachable]
                message=f"{type(self.value)} must be of type: {UserCodeStatus}"
            )
        return SyftSuccess(message=f"{type(self)} valid")

    def mutate(
        self,
        status: UserCodeStatusCollection,
        context: ChangeContext,
        undo: bool,
    ) -> UserCodeStatusCollection | SyftError:
        reason: str = context.extra_kwargs.get("reason", "")

        if not undo:
            res = status.mutate(
                value=(self.value, reason),
                server_name=context.server.name,
                server_id=context.server.id,
                verify_key=context.server.signing_key.verify_key,
            )
            if isinstance(res, SyftError):
                return res
        else:
            res = status.mutate(
                value=(UserCodeStatus.DENIED, reason),
                server_name=context.server.name,
                server_id=context.server.id,
                verify_key=context.server.signing_key.verify_key,
            )
        return res

    def _run(
        self, context: ChangeContext, apply: bool
    ) -> Result[SyftSuccess, SyftError]:
        try:
            valid = self.valid
            if not valid:
                return Err(valid)
            user_code = self.linked_user_code.resolve_with_context(context)
            if user_code.is_err():
                return Err(SyftError(message=user_code.err()))
            user_code = user_code.ok()
            user_code_status = self.linked_obj.resolve_with_context(context)
            if user_code_status.is_err():
                return Err(SyftError(message=user_code_status.err()))
            user_code_status = user_code_status.ok()

            if apply:
                # Only mutate, does not write to stash
                updated_status = self.mutate(user_code_status, context, undo=False)

                if isinstance(updated_status, SyftError):
                    return Err(updated_status.message)

                self.linked_obj.update_with_context(context, updated_status)
            else:
                updated_status = self.mutate(user_code_status, context, undo=True)
                if isinstance(updated_status, SyftError):
                    return Err(updated_status.message)

                self.linked_obj.update_with_context(context, updated_status)
            return Ok(SyftSuccess(message=f"{type(self)} Success"))
        except Exception as e:
            logger.error(f"failed to apply {type(self)}", exc_info=e)
            return Err(SyftError(message=str(e)))

    def apply(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=True)

    def undo(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=False)

    @property
    def link(self) -> SyftObject | None:
        if self.linked_obj:
            return self.linked_obj.resolve
        return None


@serializable()
class SyncedUserCodeStatusChange(UserCodeStatusChange):
    __canonical_name__ = "SyncedUserCodeStatusChange"
    __version__ = SYFT_OBJECT_VERSION_1
    linked_obj: LinkedObject | None = None  # type: ignore

    @property
    def approved(self) -> bool:
        return self.code.status.approved

    def mutate(
        self,
        status: UserCodeStatusCollection,
        context: ChangeContext,
        undo: bool,
    ) -> UserCodeStatusCollection | SyftError:
        return SyftError(
            message="Synced UserCodes status is computed, and cannot be updated manually."
        )

    def _run(
        self, context: ChangeContext, apply: bool
    ) -> Result[SyftSuccess, SyftError]:
        return Ok(
            SyftError(
                message="Synced UserCodes status is computed, and cannot be updated manually."
            )
        )

    def link(self) -> Any:  # type: ignore
        return self.code.status



# stdlib
from enum import Enum
from enum import auto

# relative
from ...serde.serializable import serializable


@serializable(canonical_name="EMAIL_TYPES", version=1)
class EMAIL_TYPES(Enum):
    PASSWORD_RESET_EMAIL = "PasswordResetTemplate"  # nosec
    ONBOARD_EMAIL = "OnBoardEmailTemplate"
    REQUEST_EMAIL = "RequestEmailTemplate"
    REQUEST_UPDATE_EMAIL = "RequestUpdateEmailTemplate"


@serializable(canonical_name="NOTIFIERS", version=1)
class NOTIFIERS(Enum):
    EMAIL = auto()
    SMS = auto()
    SLACK = auto()
    APP = auto()



# stdlib

# third party

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...store.linked_obj import LinkedObject
from ...types.uid import UID
from ...util.telemetry import instrument
from ..context import AuthedServiceContext
from ..notification.email_templates import RequestEmailTemplate
from ..notification.email_templates import RequestUpdateEmailTemplate
from ..notification.notification_service import CreateNotification
from ..notification.notification_service import NotificationService
from ..notification.notifications import Notification
from ..notifier.notifier_enums import NOTIFIERS
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user import UserView
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..user.user_service import UserService
from .request import Change
from .request import Request
from .request import RequestInfo
from .request import RequestInfoFilter
from .request import RequestStatus
from .request import SubmitRequest
from .request_stash import RequestStash


@instrument
@serializable(canonical_name="RequestService", version=1)
class RequestService(AbstractService):
    store: DocumentStore
    stash: RequestStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = RequestStash(store=store)

    @service_method(path="request.submit", name="submit", roles=GUEST_ROLE_LEVEL)
    def submit(
        self,
        context: AuthedServiceContext,
        request: SubmitRequest,
        send_message: bool = True,
        reason: str | None = "",
    ) -> Request | SyftError:
        """Submit a Request"""
        try:
            req = request.to(Request, context=context)
            result = self.stash.set(context.credentials, req)
            if result.is_ok():
                request = result.ok()
                link = LinkedObject.with_context(request, context=context)

                admin_verify_key = context.server.get_service_method(
                    UserService.admin_verify_key
                )

                root_verify_key = admin_verify_key()
                if send_message:
                    subject_msg = f"Result to request {str(request.id)[:4]}...{str(request.id)[-3:]}\
                        has been successfully deposited."
                    message = CreateNotification(
                        subject=subject_msg if not reason else reason,
                        from_user_verify_key=context.credentials,
                        to_user_verify_key=root_verify_key,
                        linked_obj=link,
                        notifier_types=[NOTIFIERS.EMAIL],
                        email_template=RequestEmailTemplate,
                    )
                    method = context.server.get_service_method(NotificationService.send)
                    result = method(context=context, notification=message)
                    if isinstance(result, Notification):
                        return request
                    else:
                        return SyftError(
                            message=f"Failed to send notification: {result.err()}"
                        )

                return request

            if result.is_err():
                return SyftError(message=str(result.err()))
            return result.ok()
        except Exception as e:
            print("Failed to submit Request", e)
            raise e

    @service_method(
        path="request.get_by_uid", name="get_by_uid", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def get_by_uid(
        self, context: AuthedServiceContext, uid: UID
    ) -> Request | None | SyftError:
        result = self.stash.get_by_uid(context.credentials, uid)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return result.ok()

    @service_method(
        path="request.get_all", name="get_all", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def get_all(self, context: AuthedServiceContext) -> list[Request] | SyftError:
        result = self.stash.get_all(context.credentials)
        if result.is_err():
            return SyftError(message=str(result.err()))
        requests = result.ok()
        # return [self.resolve_nested_requests(context, request) for request in requests]
        return requests

    @service_method(path="request.get_all_info", name="get_all_info")
    def get_all_info(
        self,
        context: AuthedServiceContext,
        page_index: int | None = 0,
        page_size: int | None = 0,
    ) -> list[list[RequestInfo]] | list[RequestInfo] | SyftError:
        """Get the information of all requests"""

        result = self.stash.get_all(context.credentials)
        if result.is_err():
            return SyftError(message=result.err())

        method = context.server.get_service_method(UserService.get_by_verify_key)
        get_message = context.server.get_service_method(
            NotificationService.filter_by_obj
        )

        requests: list[RequestInfo] = []
        for req in result.ok():
            user = method(req.requesting_user_verify_key).to(UserView)
            message = get_message(context=context, obj_uid=req.id)
            requests.append(RequestInfo(user=user, request=req, notification=message))
        if not page_size:
            return requests

        # If chunk size is defined, then split list into evenly sized chunks
        chunked_requests: list[list[RequestInfo]] = [
            requests[i : i + page_size] for i in range(0, len(requests), page_size)
        ]
        if page_index:
            return chunked_requests[page_index]
        else:
            return chunked_requests

    @service_method(path="request.add_changes", name="add_changes")
    def add_changes(
        self, context: AuthedServiceContext, uid: UID, changes: list[Change]
    ) -> Request | SyftError:
        result = self.stash.get_by_uid(credentials=context.credentials, uid=uid)

        if result.is_err():
            return SyftError(
                message=f"Failed to retrieve request with uid: {uid}. Error: {result.err()}"
            )

        request = result.ok()
        request.changes.extend(changes)
        return self.save(context=context, request=request)

    @service_method(path="request.filter_all_info", name="filter_all_info")
    def filter_all_info(
        self,
        context: AuthedServiceContext,
        request_filter: RequestInfoFilter,
        page_index: int | None = 0,
        page_size: int | None = 0,
    ) -> list[RequestInfo] | SyftError:
        """Get a Dataset"""
        result = self.get_all_info(context)
        requests = list(
            filter(lambda res: (request_filter.name in res.user.name), result)
        )

        # If chunk size is defined, then split list into evenly sized chunks
        if page_size:
            requests = [
                requests[i : i + page_size] for i in range(0, len(requests), page_size)
            ]
            if page_index is not None:
                # Return the proper slice using chunk_index
                requests = requests[page_index]

        return requests

    @service_method(
        path="request.apply",
        name="apply",
    )
    def apply(
        self,
        context: AuthedServiceContext,
        uid: UID,
        **kwargs: dict,
    ) -> SyftSuccess | SyftError:
        request = self.stash.get_by_uid(context.credentials, uid)
        if request.is_ok():
            request = request.ok()

            context.extra_kwargs = kwargs
            result = request.apply(context=context)

            filter_by_obj = context.server.get_service_method(
                NotificationService.filter_by_obj
            )
            request_notification = filter_by_obj(context=context, obj_uid=uid)

            link = LinkedObject.with_context(request, context=context)
            if not request.get_status(context) == RequestStatus.PENDING:
                if request_notification is not None and not isinstance(
                    request_notification, SyftError
                ):
                    mark_as_read = context.server.get_service_method(
                        NotificationService.mark_as_read
                    )
                    mark_as_read(context=context, uid=request_notification.id)

                    notification = CreateNotification(
                        subject=f"Your request ({str(uid)[:4]}) has been approved!",
                        from_user_verify_key=context.credentials,
                        to_user_verify_key=request.requesting_user_verify_key,
                        linked_obj=link,
                        notifier_types=[NOTIFIERS.EMAIL],
                        email_template=RequestUpdateEmailTemplate,
                    )
                    send_notification = context.server.get_service_method(
                        NotificationService.send
                    )
                    send_notification(context=context, notification=notification)

            # TODO: check whereever we're return SyftError encapsulate it in Result.
            if hasattr(result, "value"):
                return result.value
            return result
        return request.value

    @service_method(path="request.undo", name="undo")
    def undo(
        self, context: AuthedServiceContext, uid: UID, reason: str
    ) -> SyftSuccess | SyftError:
        result = self.stash.get_by_uid(credentials=context.credentials, uid=uid)
        if result.is_err():
            return SyftError(
                message=f"Failed to update request: {uid} with error: {result.err()}"
            )

        request = result.ok()
        if request is None:
            return SyftError(message=f"Request with uid: {uid} does not exists.")

        context.extra_kwargs["reason"] = reason
        result = request.undo(context=context)

        if result.is_err():
            return SyftError(
                message=f"Failed to undo Request: <{uid}> with error: {result.err()}"
            )

        link = LinkedObject.with_context(request, context=context)
        message_subject = f"Your request ({str(uid)[:4]}) has been denied. "

        notification = CreateNotification(
            subject=message_subject,
            from_user_verify_key=context.credentials,
            to_user_verify_key=request.requesting_user_verify_key,
            linked_obj=link,
            notifier_types=[NOTIFIERS.EMAIL],
            email_template=RequestUpdateEmailTemplate,
        )

        send_notification = context.server.get_service_method(NotificationService.send)
        send_notification(context=context, notification=notification)

        return SyftSuccess(message=f"Request {uid} successfully denied !")

    def save(
        self, context: AuthedServiceContext, request: Request
    ) -> Request | SyftError:
        result = self.stash.update(context.credentials, request)
        if result.is_ok():
            return result.ok()
        return SyftError(
            message=f"Failed to update Request: <{request.id}>. Error: {result.err()}"
        )

    @service_method(
        path="request.delete_by_uid",
        name="delete_by_uid",
    )
    def delete_by_uid(
        self, context: AuthedServiceContext, uid: UID
    ) -> SyftSuccess | SyftError:
        """Delete the request with the given uid."""
        result = self.stash.delete_by_uid(context.credentials, uid)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return SyftSuccess(message=f"Request with id {uid} deleted.")

    @service_method(
        path="request.set_tags",
        name="set_tags",
        roles=ADMIN_ROLE_LEVEL,
    )
    def set_tags(
        self,
        context: AuthedServiceContext,
        request: Request,
        tags: list[str],
    ) -> Request | SyftError:
        request = self.stash.get_by_uid(context.credentials, request.id)
        if request.is_err():
            return SyftError(message=str(request.err()))
        if request.ok() is None:
            return SyftError(message="Request does not exist.")
        request = request.ok()

        request.tags = tags
        return self.save(context, request)

    @service_method(path="request.get_by_usercode_id", name="get_by_usercode_id")
    def get_by_usercode_id(
        self, context: AuthedServiceContext, usercode_id: UID
    ) -> list[Request] | SyftError:
        result = self.stash.get_by_usercode_id(context.credentials, usercode_id)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return result.ok()


TYPE_TO_SERVICE[Request] = RequestService
SERVICE_TO_TYPES[RequestService].update({Request})



# stdlib
from typing import Any

# relative
from ...util.util import get_mb_serialized_size
from ..metadata.server_metadata import ServerMetadata
from ..metadata.server_metadata import ServerMetadataJSON


def min_size_for_blob_storage_upload(
    metadata: ServerMetadata | ServerMetadataJSON,
) -> int:
    return metadata.min_size_blob_storage_mb


def can_upload_to_blob_storage(
    data: Any, metadata: ServerMetadata | ServerMetadataJSON
) -> bool:
    return get_mb_serialized_size(data) >= min_size_for_blob_storage_upload(metadata)



# stdlib
from collections.abc import Callable
from datetime import datetime
from enum import Enum
import logging
import textwrap
from typing import Any

# third party
from IPython.display import display
import itables
import markdown
import pandas as pd
from pydantic import ConfigDict
from pydantic import field_validator
from pydantic import model_validator
from result import Err
from result import Ok
from result import Result
from typing_extensions import Self

# relative
from ...client.api import APIRegistry
from ...serde.serializable import serializable
from ...store.document_store import PartitionKey
from ...types.datetime import DateTime
from ...types.dicttuple import DictTuple
from ...types.syft_object import PartialSyftObject
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.transforms import TransformContext
from ...types.transforms import generate_id
from ...types.transforms import make_set_default
from ...types.transforms import transform
from ...types.transforms import validate_url
from ...types.uid import UID
from ...util import options
from ...util.colors import ON_SURFACE_HIGHEST
from ...util.colors import SURFACE
from ...util.colors import SURFACE_SURFACE
from ...util.markdown import as_markdown_python_code
from ...util.misc_objs import MarkdownDescription
from ...util.notebook_ui.icons import Icon
from ...util.notebook_ui.styles import FONT_CSS
from ...util.notebook_ui.styles import ITABLES_CSS
from ..action.action_data_empty import ActionDataEmpty
from ..action.action_object import ActionObject
from ..data_subject.data_subject import DataSubject
from ..data_subject.data_subject import DataSubjectCreate
from ..data_subject.data_subject_service import DataSubjectService
from ..response import SyftError
from ..response import SyftException
from ..response import SyftSuccess
from ..response import SyftWarning

NamePartitionKey = PartitionKey(key="name", type_=str)
logger = logging.getLogger(__name__)


@serializable()
class Contributor(SyftObject):
    __canonical_name__ = "Contributor"
    __version__ = SYFT_OBJECT_VERSION_1

    name: str
    role: str | None = None
    email: str
    phone: str | None = None
    note: str | None = None

    __repr_attrs__ = ["name", "role", "email"]

    def _repr_html_(self) -> Any:
        return f"""
            <style>
            .syft-contributor {{color: {SURFACE[options.color_theme]};}}
            </style>
            <div class='syft-contributor' style="line-height:25%">
                <h3>Contributor</h3>
                <p><strong>Name: </strong>{self.name}</p>
                <p><strong>Role: </strong>{self.role}</p>
                <p><strong>Email: </strong>{self.email}</p>
            </div>
            """

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Contributor):
            return False

        # We assoctiate two contributors as equal if they have the same email
        return self.email == value.email

    def __hash__(self) -> int:
        return hash(self.email)


@serializable()
class Asset(SyftObject):
    # version
    __canonical_name__ = "Asset"
    __version__ = SYFT_OBJECT_VERSION_1

    action_id: UID
    server_uid: UID
    name: str
    description: MarkdownDescription | None = None
    contributors: set[Contributor] = set()
    data_subjects: list[DataSubject] = []
    mock_is_real: bool = False
    shape: tuple | None = None
    created_at: DateTime = DateTime.now()
    uploader: Contributor | None = None

    # _kwarg_name and _dataset_name are set by the UserCode.assets
    _kwarg_name: str | None = None
    _dataset_name: str | None = None
    __syft_include_id_coll_repr__ = False

    def __init__(
        self,
        description: MarkdownDescription | str | None = "",
        **data: Any,
    ):
        if isinstance(description, str):
            description = MarkdownDescription(text=description)
        super().__init__(**data, description=description)

    def _repr_html_(self) -> Any:
        itables_css = f"""
        .itables table {{
            margin: 0 auto;
            float: left;
            color: {ON_SURFACE_HIGHEST[options.color_theme]};
        }}
        .itables table th {{color: {SURFACE_SURFACE[options.color_theme]};}}
        """

        # relative
        from ...service.action.action_object import ActionObject

        uploaded_by_line = (
            f"<p><strong>Uploaded by: </strong>{self.uploader.name} ({self.uploader.email})</p>"
            if self.uploader
            else ""
        )

        private_data_res = self._private_data()
        if private_data_res.is_err():
            data_table_line = private_data_res.err_value
        else:
            private_data_obj = private_data_res.ok_value
            if isinstance(private_data_obj, ActionObject):
                data_table_line = itables.to_html_datatable(
                    df=self.data.syft_action_data, css=itables_css
                )
            elif isinstance(private_data_obj, pd.DataFrame):
                data_table_line = itables.to_html_datatable(
                    df=private_data_obj, css=itables_css
                )
            else:
                data_table_line = private_data_res.ok_value

        if isinstance(self.mock, ActionObject):
            mock_table_line = itables.to_html_datatable(
                df=self.mock.syft_action_data, css=itables_css
            )
        elif isinstance(self.mock, pd.DataFrame):
            mock_table_line = itables.to_html_datatable(df=self.mock, css=itables_css)
        else:
            mock_table_line = self.mock
            if isinstance(mock_table_line, SyftError):
                mock_table_line = mock_table_line.message

        return f"""
            <style>
            {FONT_CSS}
            .syft-asset {{color: {SURFACE[options.color_theme]};}}
            .syft-asset h3,
            .syft-asset p
              {{font-family: 'Open Sans'}}
            {ITABLES_CSS}
            </style>

            <div class="syft-asset">
            <h3>{self.name}</h3>
            <p>{self.description or ""}</p>
            <p><strong>Asset ID: </strong>{self.id}</p>
            <p><strong>Action Object ID: </strong>{self.action_id}</p>
            {uploaded_by_line}
            <p><strong>Created on: </strong>{self.created_at}</p>
            <p><strong>Data:</strong></p>
            {data_table_line}
            <p><strong>Mock Data:</strong></p>
            {mock_table_line}
            </div>"""

    def __repr__(self) -> str:
        return f"Asset(name='{self.name}', server_uid='{self.server_uid}', action_id='{self.action_id}')"

    def _repr_markdown_(self, wrap_as_python: bool = True, indent: int = 0) -> str:
        _repr_str = f"Asset: {self.name}\n"
        _repr_str += f"Pointer Id: {self.action_id}\n"
        _repr_str += f"Description: {self.description}\n"
        _repr_str += f"Total Data Subjects: {len(self.data_subjects)}\n"
        _repr_str += f"Shape: {self.shape}\n"
        _repr_str += f"Contributors: {len(self.contributors)}\n"
        for contributor in self.contributors:
            _repr_str += f"\t{contributor.name}: {contributor.email}\n"
        return as_markdown_python_code(_repr_str)

    def _coll_repr_(self) -> dict[str, Any]:
        base_dict = {
            "Parameter": self._kwarg_name,
            "Action ID": self.action_id,
            "Asset Name": self.name,
            "Dataset Name": self._dataset_name,
            "Server UID": self.server_uid,
        }

        # _kwarg_name and _dataset_name are set by the UserCode.assets
        # if they are None, we remove them from the dict
        filtered_dict = {
            key: value for key, value in base_dict.items() if value is not None
        }
        return filtered_dict

    def _get_dict_for_user_code_repr(self) -> dict[str, Any]:
        return {
            "action_id": self.action_id.no_dash,
            "source_asset": self.name,
            "source_dataset": self._dataset_name,
            "source_server": self.server_uid.no_dash,
        }

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Asset):
            return False
        return (
            self.action_id == other.action_id
            and self.name == other.name
            and self.contributors == other.contributors
            and self.shape == other.shape
            and self.description == other.description
            and self.data_subjects == other.data_subjects
            and self.mock_is_real == other.mock_is_real
            and self.uploader == other.uploader
            and self.created_at == other.created_at
        )

    @property
    def pointer(self) -> Any:
        api = APIRegistry.api_for(
            server_uid=self.server_uid,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is not None and api.services is not None:
            return api.services.action.get_pointer(self.action_id)

    @property
    def mock(self) -> SyftError | Any:
        api = APIRegistry.api_for(
            server_uid=self.server_uid,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(message=f"You must login to {self.server_uid}")
        result = api.services.action.get_mock(self.action_id)
        if isinstance(result, SyftError):
            return result
        try:
            if isinstance(result, SyftObject):
                return result.syft_action_data
            return result
        except Exception as e:
            return SyftError(message=f"Failed to get mock. {e}")

    def has_data_permission(self) -> bool:
        return self.data is not None

    def has_permission(self, data_result: Any) -> bool:
        # TODO: implement in a better way
        return not (
            isinstance(data_result, str)
            and data_result.startswith("Permission")
            and data_result.endswith("denied")
        )

    def _private_data(self) -> Result[Any, str]:
        """
        Retrieves the private data associated with this asset.

        Returns:
            Result[Any, str]: A Result object containing the private data if the user has permission
            otherwise an Err object with the message "You do not have permission to access private data."
        """
        api = APIRegistry.api_for(
            server_uid=self.server_uid,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None or api.services is None:
            return Ok(None)
        res = api.services.action.get(self.action_id)
        if self.has_permission(res):
            return Ok(res.syft_action_data)
        else:
            return Err("You do not have permission to access private data.")

    @property
    def data(self) -> Any:
        # relative
        private_data_or_error = self._private_data()

        if private_data_or_error.is_err():
            display(SyftError(message=private_data_or_error.err_value), clear=True)
            return None
        else:
            return private_data_or_error.ok_value


def _is_action_data_empty(obj: Any) -> bool:
    # just a wrapper of action_object.is_action_data_empty
    # to work around circular import error

    # relative
    from ...service.action.action_object import is_action_data_empty

    return is_action_data_empty(obj)


def check_mock(data: Any, mock: Any) -> bool:
    if type(data) == type(mock):
        return True

    return _is_action_data_empty(mock) or _is_action_data_empty(data)


@serializable()
class CreateAsset(SyftObject):
    # version
    __canonical_name__ = "CreateAsset"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID | None = None  # type:ignore[assignment]
    name: str
    description: MarkdownDescription | None = None
    contributors: set[Contributor] = set()
    data_subjects: list[DataSubjectCreate] = []
    server_uid: UID | None = None
    action_id: UID | None = None
    data: Any | None = None
    mock: Any | None = None
    shape: tuple | None = None
    mock_is_real: bool = False
    created_at: DateTime | None = None
    uploader: Contributor | None = None

    __repr_attrs__ = ["name"]
    model_config = ConfigDict(validate_assignment=True, extra="forbid")

    def __init__(self, description: str | None = None, **data: Any) -> None:
        if isinstance(description, str):
            description = MarkdownDescription(text=description)
        super().__init__(**data, description=description)

    @model_validator(mode="after")
    def __mock_is_real_for_empty_mock_must_be_false(self) -> Self:
        if self.mock_is_real and (
            self.mock is None or _is_action_data_empty(self.mock)
        ):
            self.__dict__["mock_is_real"] = False

        return self

    def contains_empty(self) -> bool:
        if isinstance(self.mock, ActionObject) and isinstance(
            self.mock.syft_action_data_cache, ActionDataEmpty
        ):
            return True
        if isinstance(self.data, ActionObject) and isinstance(
            self.data.syft_action_data_cache, ActionDataEmpty
        ):
            return True
        return False

    def add_data_subject(self, data_subject: DataSubject) -> None:
        self.data_subjects.append(data_subject)

    def add_contributor(
        self,
        name: str,
        email: str,
        role: Enum | str | None = None,
        phone: str | None = None,
        note: str | None = None,
    ) -> SyftSuccess | SyftError:
        try:
            _role_str = role.value if isinstance(role, Enum) else role
            contributor = Contributor(
                name=name, role=_role_str, email=email, phone=phone, note=note
            )
            if contributor in self.contributors:
                return SyftError(
                    message=f"Contributor with email: '{email}' already exists in '{self.name}' Asset."
                )
            self.contributors.add(contributor)

            return SyftSuccess(
                message=f"Contributor '{name}' added to '{self.name}' Asset."
            )
        except Exception as e:
            return SyftError(message=f"Failed to add contributor. Error: {e}")

    def set_description(self, description: str) -> None:
        self.description = MarkdownDescription(text=description)

    def set_obj(self, data: Any) -> None:
        if isinstance(data, SyftError):
            raise SyftException(data)
        self.data = data

    def set_mock(self, mock_data: Any, mock_is_real: bool) -> None:
        if isinstance(mock_data, SyftError):
            raise SyftException(mock_data)

        if mock_is_real and (mock_data is None or _is_action_data_empty(mock_data)):
            raise SyftException("`mock_is_real` must be False if mock is empty")

        self.mock = mock_data
        self.mock_is_real = mock_is_real

    def no_mock(self) -> None:
        # relative
        from ..action.action_object import ActionObject

        self.set_mock(ActionObject.empty(), False)

    def set_shape(self, shape: tuple) -> None:
        self.shape = shape

    def check(self) -> SyftSuccess | SyftError:
        if not check_mock(self.data, self.mock):
            return SyftError(
                message=f"set_obj type {type(self.data)} must match set_mock type {type(self.mock)}"
            )
        # if not _is_action_data_empty(self.mock):
        #     data_shape = get_shape_or_len(self.data)
        #     mock_shape = get_shape_or_len(self.mock)
        #     if data_shape != mock_shape:
        #         return SyftError(
        #             message=f"set_obj shape {data_shape} must match set_mock shape {mock_shape}"
        #         )

        return SyftSuccess(message="Dataset is Valid")


def get_shape_or_len(obj: Any) -> tuple[int, ...] | int | None:
    if hasattr(obj, "shape"):
        shape = getattr(obj, "shape", None)
        if shape:
            return shape
    len_attr = getattr(obj, "__len__", None)
    if len_attr is not None:
        len_value = len_attr()
        if isinstance(len_value, int):
            return (len_value,)
        return len_value
    return None


@serializable()
class Dataset(SyftObject):
    # version
    __canonical_name__: str = "Dataset"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    name: str
    server_uid: UID | None = None
    asset_list: list[Asset] = []
    contributors: set[Contributor] = set()
    citation: str | None = None
    url: str | None = None
    description: MarkdownDescription | None = None
    updated_at: str | None = None
    requests: int | None = 0
    mb_size: float | None = None
    created_at: DateTime = DateTime.now()
    uploader: Contributor
    summary: str | None = None
    to_be_deleted: bool = False

    __attr_searchable__ = [
        "name",
        "citation",
        "url",
        "description",
        "action_ids",
        "summary",
    ]
    __attr_unique__ = ["name"]
    __repr_attrs__ = ["name", "summary", "url", "created_at"]
    __table_sort_attr__ = "Created at"

    def __init__(
        self,
        description: str | MarkdownDescription | None = "",
        **data: Any,
    ) -> None:
        if isinstance(description, str):
            description = MarkdownDescription(text=description)
        super().__init__(**data, description=description)

    @property
    def icon(self) -> str:
        return Icon.FOLDER.svg

    def _coll_repr_(self) -> dict[str, Any]:
        return {
            "Name": self.name,
            "Summary": self.summary,
            "Assets": len(self.asset_list),
            "Size": f"{self.mb_size} (MB)",
            "Url": self.url,
            "Created at": str(self.created_at),
        }

    def _repr_html_(self) -> Any:
        uploaded_by_line = (
            (
                "<p class='paragraph-sm'><strong>"
                + f"<span class='pr-8'>Uploaded by: </span></strong>{self.uploader.name} ({self.uploader.email})</p>"
            )
            if self.uploader
            else ""
        )
        if self.description is not None and self.description.text:
            description_info_message = f"""
            <h2><strong><span class='pr-8'>Description</span></strong></h2>
            {markdown.markdown(self.description.text, extensions=["extra"])}
            """
        else:
            description_info_message = ""
        if self.to_be_deleted:
            return "This dataset has been marked for deletion. The underlying data may be not available."
        return f"""
            <style>
            {FONT_CSS}
            .syft-dataset {{color: {SURFACE[options.color_theme]};}}
            .syft-dataset h3,
            .syft-dataset p
              {{font-family: 'Open Sans';}}
              {ITABLES_CSS}
            </style>
            <div class='syft-dataset'>
            <h1>{self.name}</h1>
            <h2><strong><span class='pr-8'>Summary</span></strong></h2>
            {f"<p>{self.summary}</p>" if self.summary else ""}
            {description_info_message}
            <h2><strong><span class='pr-8'>Dataset Details</span></strong></h2>
            {uploaded_by_line}
            <p class='paragraph-sm'><strong><span class='pr-8'>Created on: </span></strong>{self.created_at}</p>
            <p class='paragraph-sm'><strong><span class='pr-8'>URL:
            </span></strong><a href='{self.url}'>{self.url}</a></p>
            <p class='paragraph-sm'><strong><span class='pr-8'>Contributors:</span></strong>
            To see full details call <strong>dataset.contributors</strong>.</p>
            <h2><strong><span class='pr-8'>Assets</span></strong></h2>
            {self.assets._repr_html_()}
            """

    def action_ids(self) -> list[UID]:
        return [asset.action_id for asset in self.asset_list if asset.action_id]

    @property
    def assets(self) -> DictTuple[str, Asset]:
        return DictTuple((asset.name, asset) for asset in self.asset_list)

    def _repr_markdown_(self, wrap_as_python: bool = True, indent: int = 0) -> str:
        _repr_str = f"Syft Dataset: {self.name}\n\n"
        _repr_str += "Assets:\n\n"
        for asset in self.asset_list:
            if asset.description is not None:
                description_text = textwrap.shorten(
                    asset.description.text, width=100, placeholder="..."
                )
                _repr_str += f"\t{asset.name}: {description_text}\n\n"
            else:
                _repr_str += f"\t{asset.name}\n\n"
        if self.citation:
            _repr_str += f"Citation: {self.citation}\n\n"
        if self.url:
            _repr_str += f"URL: {self.url}\n\n"
        if self.description:
            _repr_str += f"Description: \n\n{self.description.text}\n\n"
        return _repr_str

    @property
    def client(self) -> Any | None:
        # relative
        from ...client.client import SyftClientSessionCache

        client = SyftClientSessionCache.get_client_for_server_uid(self.server_uid)
        if client is None:
            return SyftError(
                message=f"No clients for {self.server_uid} in memory. Please login with sy.login"
            )
        return client


_ASSET_WITH_NONE_MOCK_ERROR_MESSAGE: str = "".join(
    [
        "To be included in a Dataset, an asset must either contain a mock, ",
        "or have it explicitly set to be empty.\n",
        "You can create an asset without a mock with `sy.Asset(..., mock=sy.ActionObject.empty())` or\n"
        "set the mock of an existing asset to be empty with `asset.no_mock()` or ",
        "`asset.mock = sy.ActionObject.empty()`.",
    ]
)


def _check_asset_must_contain_mock(asset_list: list[CreateAsset]) -> None:
    assets_without_mock = [asset.name for asset in asset_list if asset.mock is None]
    if assets_without_mock:
        raise ValueError(
            "".join(
                [
                    "These assets do not contain a mock:\n",
                    *[f"{asset}\n" for asset in assets_without_mock],
                    "\n",
                    _ASSET_WITH_NONE_MOCK_ERROR_MESSAGE,
                ]
            )
        )


@serializable()
class DatasetPageView(SyftObject):
    # version
    __canonical_name__ = "DatasetPageView"
    __version__ = SYFT_OBJECT_VERSION_1

    datasets: DictTuple
    total: int


@serializable()
class CreateDataset(Dataset):
    # version
    __canonical_name__ = "CreateDataset"
    __version__ = SYFT_OBJECT_VERSION_1
    asset_list: list[CreateAsset] = []

    __repr_attrs__ = ["name", "summary", "url"]

    id: UID | None = None  # type: ignore[assignment]
    created_at: DateTime | None = None  # type: ignore[assignment]
    uploader: Contributor | None = None  # type: ignore[assignment]

    model_config = ConfigDict(validate_assignment=True, extra="forbid")

    @field_validator("asset_list")
    @classmethod
    def __assets_must_contain_mock(
        cls, asset_list: list[CreateAsset]
    ) -> list[CreateAsset]:
        _check_asset_must_contain_mock(asset_list)
        return asset_list

    @field_validator("to_be_deleted")
    @classmethod
    def __to_be_deleted_must_be_false(cls, v: bool) -> bool:
        if v is True:
            raise ValueError("to_be_deleted must be False")
        return v

    def set_description(self, description: str) -> None:
        self.description = MarkdownDescription(text=description)

    def set_summary(self, summary: str) -> None:
        self.summary = summary

    def add_citation(self, citation: str) -> None:
        self.citation = citation

    def add_url(self, url: str) -> None:
        self.url = url

    def add_contributor(
        self,
        name: str,
        email: str,
        role: Enum | str | None = None,
        phone: str | None = None,
        note: str | None = None,
    ) -> SyftSuccess | SyftError:
        try:
            _role_str = role.value if isinstance(role, Enum) else role
            contributor = Contributor(
                name=name, role=_role_str, email=email, phone=phone, note=note
            )
            if contributor in self.contributors:
                return SyftError(
                    message=f"Contributor with email: '{email}' already exists in '{self.name}' Dataset."
                )
            self.contributors.add(contributor)
            return SyftSuccess(
                message=f"Contributor '{name}' added to '{self.name}' Dataset."
            )
        except Exception as e:
            return SyftError(message=f"Failed to add contributor. Error: {e}")

    def add_asset(
        self, asset: CreateAsset, force_replace: bool = False
    ) -> SyftSuccess | SyftError:
        if asset.mock is None:
            raise ValueError(_ASSET_WITH_NONE_MOCK_ERROR_MESSAGE)

        for i, existing_asset in enumerate(self.asset_list):
            if existing_asset.name == asset.name:
                if not force_replace:
                    return SyftError(
                        message=f"""Asset "{asset.name}" already exists in '{self.name}' Dataset."""
                        """ Use add_asset(asset, force_replace=True) to replace."""
                    )
                else:
                    self.asset_list[i] = asset
                    return SyftSuccess(
                        message=f"Asset {asset.name} has been successfully replaced."
                    )

        self.asset_list.append(asset)

        return SyftSuccess(
            message=f"Asset '{asset.name}' added to '{self.name}' Dataset."
        )

    def replace_asset(self, asset: CreateAsset) -> SyftSuccess | SyftError:
        return self.add_asset(asset=asset, force_replace=True)

    def remove_asset(self, name: str) -> SyftSuccess | SyftError:
        asset_to_remove = None
        for asset in self.asset_list:
            if asset.name == name:
                asset_to_remove = asset
                break

        if asset_to_remove is None:
            return SyftError(message=f"No asset exists with name: {name}")
        self.asset_list.remove(asset_to_remove)
        return SyftSuccess(
            message=f"Asset '{self.name}' removed from '{self.name}' Dataset."
        )

    def check(self) -> Result[SyftSuccess, list[SyftError]]:
        errors = []
        for asset in self.asset_list:
            result = asset.check()
            if not result:
                errors.append(result)
        if len(errors):
            return Err(errors)
        return Ok(SyftSuccess(message="Dataset is Valid"))


def create_and_store_twin(context: TransformContext) -> TransformContext:
    if context.output is None:
        raise ValueError(f"{context}'s output is None. No transformation happened")

    action_id = context.output["action_id"]
    if action_id is None:
        # relative
        from ...types.twin_object import TwinObject

        private_obj = context.output.pop("data", None)
        mock_obj = context.output.pop("mock", None)
        if private_obj is None and mock_obj is None:
            raise ValueError("No data and no action_id means this asset has no data")

        asset = context.obj  # type: ignore
        contains_empty = asset.contains_empty()  # type: ignore
        twin = TwinObject(
            private_obj=asset.data,  # type: ignore
            mock_obj=asset.mock,  # type: ignore
            syft_server_location=asset.syft_server_location,  # type: ignore
            syft_client_verify_key=asset.syft_client_verify_key,  # type: ignore
        )
        res = twin._save_to_blob_storage(allow_empty=contains_empty)
        if isinstance(res, SyftError):
            raise ValueError(res.message)
        if isinstance(res, SyftWarning):
            logger.debug(res.message)
        # TODO, upload to blob storage here
        if context.server is None:
            raise ValueError(
                "f{context}'s server is None, please log in. No trasformation happened"
            )
        action_service = context.server.get_service("actionservice")
        result = action_service._set(
            context=context.to_server_context(),
            action_object=twin,
        )
        if result.is_err():
            raise RuntimeError(f"Failed to create and store twin. Error: {result}")

        context.output["action_id"] = twin.id
    else:
        private_obj = context.output.pop("data", None)
        mock_obj = context.output.pop("mock", None)

    return context


def infer_shape(context: TransformContext) -> TransformContext:
    if context.output is None:
        raise ValueError(f"{context}'s output is None. No transformation happened")
    if context.output["shape"] is None:
        if context.obj is not None and not _is_action_data_empty(context.obj.mock):
            context.output["shape"] = get_shape_or_len(context.obj.mock)
    return context


def set_data_subjects(context: TransformContext) -> TransformContext | SyftError:
    if context.output is None:
        raise ValueError(f"{context}'s output is None. No transformation happened")
    if context.server is None:
        return SyftError(
            "f{context}'s server is None, please log in. No trasformation happened"
        )
    data_subjects = context.output["data_subjects"]
    get_data_subject = context.server.get_service_method(DataSubjectService.get_by_name)
    resultant_data_subjects = []
    for data_subject in data_subjects:
        result = get_data_subject(context=context, name=data_subject.name)
        if isinstance(result, SyftError):
            return result
        resultant_data_subjects.append(result)
    context.output["data_subjects"] = resultant_data_subjects
    return context


def add_msg_creation_time(context: TransformContext) -> TransformContext:
    if context.output is None:
        return context

    context.output["created_at"] = DateTime.now()
    return context


def add_default_server_uid(context: TransformContext) -> TransformContext:
    if context.output is not None:
        if context.output["server_uid"] is None and context.server is not None:
            context.output["server_uid"] = context.server.id
    else:
        raise ValueError(f"{context}'s output is None. No transformation happened")
    return context


@transform(CreateAsset, Asset)
def createasset_to_asset() -> list[Callable]:
    return [
        generate_id,
        add_msg_creation_time,
        infer_shape,
        create_and_store_twin,
        set_data_subjects,
        add_default_server_uid,
    ]


def convert_asset(context: TransformContext) -> TransformContext:
    if context.output is None:
        return context

    assets = context.output.pop("asset_list", [])
    for idx, create_asset in enumerate(assets):
        asset_context = TransformContext.from_context(obj=create_asset, context=context)
        assets[idx] = create_asset.to(Asset, context=asset_context)
    context.output["asset_list"] = assets

    return context


def add_current_date(context: TransformContext) -> TransformContext:
    if context.output is None:
        return context

    current_date = datetime.now()
    formatted_date = current_date.strftime("%b %d, %Y")
    context.output["updated_at"] = formatted_date

    return context


@transform(CreateDataset, Dataset)
def createdataset_to_dataset() -> list[Callable]:
    return [
        generate_id,
        add_msg_creation_time,
        validate_url,
        convert_asset,
        add_current_date,
        make_set_default("to_be_deleted", False),  # explicitly set it to False
    ]


class DatasetUpdate(PartialSyftObject):
    __canonical_name__ = "DatasetUpdate"
    __version__ = SYFT_OBJECT_VERSION_1

    name: str
    to_be_deleted: bool



# future
from __future__ import annotations

# stdlib
import os
from pathlib import Path
import shutil
import tempfile

# relative
from ..types.uid import UID


def get_named_server_uid(name: str) -> UID:
    """
    Get a unique identifier for a named server.
    """
    return UID.with_seed(name)


def get_temp_dir_for_server(server_uid: UID, dir_name: str = "") -> Path:
    """
    Get a temporary directory unique to the server.
    Provide all dbs, blob dirs, and locks using this directory.
    """
    root = os.getenv("SYFT_TEMP_ROOT", "syft")
    p = Path(tempfile.gettempdir(), root, str(server_uid), dir_name)
    p.mkdir(parents=True, exist_ok=True)
    return p


def remove_temp_dir_for_server(server_uid: UID) -> None:
    """
    Remove the temporary directory for this server.
    """
    rootdir = get_temp_dir_for_server(server_uid)
    if rootdir.exists():
        shutil.rmtree(rootdir, ignore_errors=True)



# stdlib
import datetime
from typing import Any

# third party
from pydantic import model_validator

# relative
from ....client.sync_decision import SyncDirection
from ....service.code.user_code import UserCode
from ....service.job.job_stash import Job
from ....service.request.request import Request
from ....service.response import SyftError
from ....service.user.user import UserView
from ....types.datetime import DateTime
from ....types.datetime import format_timedelta_human_readable
from ....types.syft_object import SYFT_OBJECT_VERSION_1
from ....types.syft_object import SyftObject
from ..icons import Icon
from ..styles import CSS_CODE
from .base import HTMLComponentBase

COPY_CSS = """
.copy-container {
  cursor: pointer;
  border-radius: 3px;
  padding: 0px 3px;
  display: inline-block;
  transition: background-color 0.3s;
  user-select: none;
  color: #B4B0BF;
  overflow: hidden;
  white-space: nowrap;
  vertical-align: middle;
}

.copy-container:hover {
  background-color: #f5f5f5;
}

.copy-container:active {
  background-color: #ebebeb;
}

.copy-text-display {
  display: inline-block;
  max-width: 50px;
  white-space: nowrap;
  overflow: hidden;
  text-overflow: ellipsis;
  vertical-align: bottom;
}
"""


class CopyButton(HTMLComponentBase):
    __canonical_name__ = "CopyButton"
    __version__ = SYFT_OBJECT_VERSION_1
    copy_text: str
    max_width: int = 50

    def format_copy_text(self, copy_text: str) -> str:
        return copy_text

    def to_html(self) -> str:
        copy_js = f"event.stopPropagation(); navigator.clipboard.writeText('{self.copy_text}');"
        text_formatted = self.format_copy_text(self.copy_text)
        button_html = f"""
        <style>{COPY_CSS}</style>
        <div class="copy-container" onclick="{copy_js}">
            <span class="copy-text-display" style="max-width: {self.max_width}px;">
                {text_formatted}
            </span>
            {Icon.COPY.svg}
        </div>
        """
        return button_html


class CopyIDButton(CopyButton):
    __canonical_name__ = "CopyIDButton"
    __version__ = SYFT_OBJECT_VERSION_1

    def format_copy_text(self, copy_text: str) -> str:
        return f"#{copy_text}"


class SyncTableObject(HTMLComponentBase):
    __canonical_name__ = "SyncTableObject"
    __version__ = SYFT_OBJECT_VERSION_1

    object: SyftObject

    def get_status_str(self) -> str:
        if isinstance(self.object, UserCode):
            return ""
        elif isinstance(self.object, Job):  # type: ignore
            return f"Status: {self.object.status.value}"
        elif isinstance(self.object, Request):
            code = self.object.code
            statusses = list(code.status.status_dict.values())
            if len(statusses) != 1:
                raise ValueError("Request code should have exactly one status")
            status_tuple = statusses[0]
            status, _ = status_tuple
            return status.value
        return ""  # type: ignore

    def get_updated_by(self) -> str:
        # TODO replace with centralized SyftObject created/updated by attribute
        if isinstance(self.object, Request):
            email = self.object.requesting_user_email
            if email is not None:
                return f"Requested by {email}"

        user_view: UserView | SyftError | None = None
        if isinstance(self.object, UserCode):
            user_view = self.object.user

        if isinstance(user_view, UserView):
            return f"Created by {user_view.email}"
        return ""

    def get_updated_delta_str(self) -> str:
        # TODO replace with centralized SyftObject created/updated by attribute
        if isinstance(self.object, Job):
            # NOTE Job is not using DateTime for creation_time, so we need to handle it separately
            time_str = self.object.creation_time
            if time_str is not None:
                t = datetime.datetime.fromisoformat(time_str)
                delta = datetime.datetime.now(datetime.timezone.utc) - t
                return f"{format_timedelta_human_readable(delta)} ago"

        dt: DateTime | None = None
        if isinstance(self.object, Request):
            dt = self.object.request_time
        if isinstance(self.object, UserCode):
            dt = self.object.submit_time
        if dt is not None:
            delta = DateTime.now().timedelta(dt)
            delta_str = format_timedelta_human_readable(delta)
            return f"{delta_str} ago"

        return ""

    def to_html(self) -> str:
        type_html = TypeLabel(object=self.object).to_html()

        type_html = TypeLabel(object=self.object).to_html()
        description_html = MainDescription(object=self.object).to_html()
        copy_id_button = CopyIDButton(
            copy_text=str(self.object.id.id), max_width=60
        ).to_html()

        updated_delta_str = self.get_updated_delta_str()
        updated_by = self.get_updated_by()
        status_str = self.get_status_str()
        status_row = " • ".join(
            s for s in [status_str, updated_by, updated_delta_str] if s
        )
        summary_html = f"""
            <div style="display: flex; gap: 8px; justify-content: space-between; width: 100%; overflow: hidden; align-items: center;">
            <div style="display: flex; gap: 8px; justify-content: start; align-items: center;">
            {type_html} {description_html}
            </div>
            {copy_id_button}
            </div>
            <div style="display: table-row">
            <span class='syncstate-col-footer'>
            {status_row}
            </span>
            </div>
        """  # noqa: E501
        summary_html = summary_html.replace("\n", "").replace("    ", "")
        return summary_html


ALERT_CSS = """
.syft-alert-container {
    padding: 4px;
    display: flex;
    justify-content: center;
}

.syft-alert-info {
    display: flex;
    align-items: center;
    width: 100%;
    padding: 8px 10px;
    gap: 8px;
    border-radius: 4px;
    background: #C2DEF0;
    color: #1F567A;
    line-height: 1.4;
    font-size: 12px;
    font-family: 'Open Sans';
}
"""


class Alert(HTMLComponentBase):
    __canonical_name__ = "Alert"
    __version__ = SYFT_OBJECT_VERSION_1
    message: str

    def to_html(self) -> str:
        full_message = f"{Icon.INFO.svg} {self.message}"
        return f"""
            <style>{ALERT_CSS}</style>
            <div class="syft-alert-container">
            <div class="syft-alert-info">
            {full_message}
            </div>
            </div>
            """


class Badge(HTMLComponentBase):
    __canonical_name__ = "Badge"
    __version__ = SYFT_OBJECT_VERSION_1
    value: str
    badge_class: str

    def to_html(self) -> str:
        value = str(self.value).upper()
        return f'<span class="badge {self.badge_class}">{value}</span>'


class Label(HTMLComponentBase):
    __canonical_name__ = "Label"
    __version__ = SYFT_OBJECT_VERSION_1
    value: str
    label_class: str

    def to_html(self) -> str:
        value = str(self.value).upper()
        return f'<span class="label {self.label_class}">{value}</span>'


class TypeLabel(Label):
    __canonical_name__ = "TypeLabel"
    __version__ = SYFT_OBJECT_VERSION_1
    object: SyftObject

    @model_validator(mode="before")
    @classmethod
    def validate_label(cls, data: dict) -> dict:
        obj = data["object"]
        data["label_class"] = cls.type_label_class(obj)
        data["value"] = type(obj).__name__.upper()
        return data

    @staticmethod
    def type_label_class(obj: Any) -> str:
        if isinstance(obj, UserCode):
            return "label-light-blue"
        elif isinstance(obj, Job):  # type: ignore
            return "label-light-blue"
        elif isinstance(obj, Request):  # type: ignore
            # TODO: handle other requests
            return "label-light-purple"
        return "label-light-blue"  # type: ignore


class MainDescription(HTMLComponentBase):
    __canonical_name__ = "CopyButton"
    __version__ = SYFT_OBJECT_VERSION_1
    object: SyftObject

    def main_object_description_str(self) -> str:
        if isinstance(self.object, UserCode):
            return self.object.service_func_name
        elif isinstance(self.object, Job):  # type: ignore
            return self.object.user_code_name or ""
        elif isinstance(self.object, Request):  # type: ignore
            # TODO: handle other requests
            return f"Execute {self.object.code.service_func_name}"
        # SyftLog
        # ExecutionOutput
        # ActionObject
        # UserCodeStatusCollection

        return ""  # type: ignore

    def to_html(self) -> str:
        return f'<span class="syncstate-description">{self.main_object_description_str()}</span>'


class SyncWidgetHeader(SyncTableObject):
    diff_batch: Any

    @model_validator(mode="before")
    @classmethod
    def add_object(cls, values: dict) -> dict:
        if "diff_batch" not in values:
            raise ValueError("diff_batch is required")
        diff_batch = values["diff_batch"]
        values["object"] = diff_batch.root_diff.non_empty_object
        return values

    def to_html(self) -> str:
        # CSS Styles
        style = CSS_CODE

        first_line_html = "<span style='color: #B4B0BF;'>Syncing changes on</span>"

        type_html = TypeLabel(object=self.object).to_html()
        description_html = MainDescription(object=self.object).to_html()
        copy_id_button = CopyIDButton(
            copy_text=str(self.object.id.id), max_width=60
        ).to_html()

        second_line_html = f"""
            <div class="widget-header2">
            <div class="widget-header2-2">
            {type_html} {description_html}
            </div>
            {copy_id_button}
            </div>
        """  # noqa: E501

        num_diffs = len(self.diff_batch.get_dependencies(include_roots=True))
        if self.diff_batch.sync_direction == SyncDirection.HIGH_TO_LOW:
            source_side = "High"
            target_side = "Low"
        else:
            source_side = "Low"
            target_side = "High"

        # Third line HTML
        third_line_html = f"<span style='color: #5E5A72;'>This would sync <span style='color: #B8520A'>{num_diffs} changes </span> from <i>{source_side} Server</i> to <i>{target_side} Server</i></span>"  # noqa: E501

        header_html = f"""
        {style}
        {first_line_html}
        {second_line_html}
        {third_line_html}
        <div style='height: 16px;'></div>
        """

        return header_html



# stdlib
from typing import Any

# third party
from typing_extensions import Self

# relative
from ..abstract_server import AbstractServer
from ..server.credentials import SyftVerifyKey
from ..server.credentials import UserLoginCredentials
from ..types.syft_object import Context
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftBaseObject
from ..types.syft_object import SyftObject
from ..types.uid import UID
from .user.user_roles import ROLE_TO_CAPABILITIES
from .user.user_roles import ServiceRole
from .user.user_roles import ServiceRoleCapability


class ServerServiceContext(Context, SyftObject):
    __canonical_name__ = "ServerServiceContext"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID | None = None  # type: ignore[assignment]
    server: AbstractServer


class AuthedServiceContext(ServerServiceContext):
    __canonical_name__ = "AuthedServiceContext"
    __version__ = SYFT_OBJECT_VERSION_1

    credentials: SyftVerifyKey
    role: ServiceRole = ServiceRole.NONE
    job_id: UID | None = None
    extra_kwargs: dict = {}
    has_execute_permissions: bool = False
    is_blocking_api_call: bool = False

    @property
    def dev_mode(self) -> Any:
        return self.server.dev_mode  # type: ignore

    def capabilities(self) -> list[ServiceRoleCapability]:
        return ROLE_TO_CAPABILITIES.get(self.role, [])

    def with_credentials(self, credentials: SyftVerifyKey, role: ServiceRole) -> Self:
        return AuthedServiceContext(
            credentials=credentials, role=role, server=self.server
        )

    def as_root_context(self) -> Self:
        return AuthedServiceContext(
            credentials=self.server.verify_key,
            role=ServiceRole.ADMIN,
            server=self.server,
        )

    @property
    def job(self):  # type: ignore
        # TODO: fix the mypy issue. The return type is Optional[Job].
        # but we can't import Job since it's a circular import
        if self.job_id is None:
            return None
        res = self.server.job_stash.get_by_uid(self.credentials, self.job_id)
        if res.is_err():
            return None
        else:
            return res.ok()


class UnauthedServiceContext(ServerServiceContext):
    __canonical_name__ = "UnauthedServiceContext"
    __version__ = SYFT_OBJECT_VERSION_1

    login_credentials: UserLoginCredentials
    server: AbstractServer
    role: ServiceRole = ServiceRole.NONE


class ChangeContext(SyftBaseObject):
    __canonical_name__ = "ChangeContext"
    __version__ = SYFT_OBJECT_VERSION_1

    server: AbstractServer
    approving_user_credentials: SyftVerifyKey | None = None
    requesting_user_credentials: SyftVerifyKey | None = None
    extra_kwargs: dict = {}

    @classmethod
    def from_service(cls, context: AuthedServiceContext) -> Self:
        return cls(
            server=context.server,
            approving_user_credentials=context.credentials,
            extra_kwargs=context.extra_kwargs,
        )

    def to_service_ctx(self) -> AuthedServiceContext:
        return AuthedServiceContext(
            server=self.server,
            credentials=self.approving_user_credentials,
            extra_kwargs=self.extra_kwargs,
        )



# relative
from ...util.notebook_ui.styles import CSS_CODE
from ...util.notebook_ui.styles import JS_DOWNLOAD_FONTS

type_html = """
            <div class="label label-light-blue"
            style="display: flex; align-items:center; justify-content: center; width: 34px; height:21px; radius:4px;
                padding: 2px, 6px, 2px, 6px">
            <span style="font-family: DejaVu Sans Mono, sans-serif;
                font-size: 12px; font-weight: 400; line-height:16.8px">
            ${job_type}</span>
            </div>
"""

header_line_html = (
    """
    <div style="height:16px;"></div>
    <div style="gap: 12px; height: 20 px; font-family: DejaVu Sans Mono, sans-serif; font-size: 14px; font-weight: 400;
        line-height:16.8px; color: #4392C5">${api_header}</div>
    <div style="height:16px;"></div>
    <div style="display: flex; gap: 12px; justify-content: start; width: 100%; overflow:
        hidden; align-items: center;">
        <div style="display: flex; gap: 12px; justify-content: start; align-items: center;
            border: 0px, 0px, 2px, 0px; padding: 0px, 0px, 16px, 0px">"""
    + type_html
    + """
        <span class='jobs-title'>${user_code_name}</span>
        </div>
        ${button_html}
    </div>
    <div style="height:16px;"></div>
"""
)  # noqa: E501

attrs_html = """<div style="display: table-row; padding: 0px, 0px, 12px, 0px; gap:8px">
    <div style="margin-top: 6px; margin-bottom: 6px;">
    <span style="font-weight: 700; line-weight: 19.6px; font-size: 14px; font: 'Open Sans'">UserCode:</span>
        ${user_code_name}
    </div>
    <div style="margin-top: 6px; margin-bottom: 6px;">
    <span style="font-weight: 700; line-weight: 19.6px; font-size: 14px; font: 'Open Sans'">Status:</span>
        ${status}
    </div>
    <div style="margin-top: 6px; margin-bottom: 6px;">
        <span style="font-weight: 700; line-weight: 19.6px; font-size: 14px; font: 'Open Sans'">
            Started At:</span>
        ${creation_time} by ${user_repr}
    </div>
    <div style="margin-top: 6px; margin-bottom: 6px;">
        <span style="font-weight: 700; line-weight: 19.6px; font-size: 14px; font: 'Open Sans'">
        Updated At:</span>
        ${updated_at}
    </div>
    ${worker_attr}
    <div style="margin-top: 6px; margin-bottom: 6px;">
    <span style="font-weight: 700; line-weight: 19.6px; font-size: 14px; font: 'Open Sans'">Subjobs:</span>
        ${no_subjobs}
    </div>
</div>
<div style="height:16px;"></div>
"""

logs_html = """
<style>
pre {
    counter-reset: line -1;
    background-color: transparent;
    color: black;
}

pre code {
    display: block;
    counter-increment: line;
    background-color: transparent;
    color: black;
    font-family: DejaVu Sans Mono, sans-serif;
    font-size: 12px;
    font-weight: 400;
    line-height: 16.8px;
}

pre code::before {
    content: counter(line);
    display: inline-block;
    width: 2em;
    padding-right: 1.5em;
    margin-right: 1.5em;
    text-align: right;
}

pre code:first-of-type::before {
    content: "#";
    font-weight: bold;
}

.logsTab {
    color: #000000;
    background: #F4F3F6;
    border-color: #CFCDD6;
    border-width: 0.5px;
    border-style: solid;
    padding: 24px;
    gap: 8px;
    margin-top: 24px;
    display: none;
    align-items: left;

}

</style>

<div id="${logs_tab_id}" class="tab-${identifier} logsTab">
    <pre>
        ${logs_lines_html}
    </pre>
</div>
"""

# TODO: add style change for selected tab
onclick_html = """<script>
function onClick_${identifier}(evt, tabname) {
    existing_tabs = document.getElementsByClassName("tab-${identifier}");
    for (i = 0; i < existing_tabs.length; i++) {
        existing_tabs[i].style.display = "none";
    }
    tablinks = document.getElementsByClassName("tablink-${identifier}");
    for (i = 0; i < tablinks.length; i++) {
        tablinks[i].className = tablinks[i].className.replace(" active", "");
    }
    tablinks = document.getElementsByClassName("tablink-border-${identifier}");
    for (i = 0; i < tablinks.length; i++) {
        tablinks[i].className = tablinks[i].className.replace(" active-border", "");
    }
    document.getElementById(tabname).style.display = "block";
    evt.currentTarget.className += " active";
    evt.currentTarget.parentNode.className += " active-border";
}
</script>
"""


tabs_html = """
    <div style="margin-top: 8px; padding: 8px, 0px, 8px, 0px; gap: 16px">
    <div style="display: flex;">
        <div class="tablink-border-${identifier} log-tab-header active-border">
            <a  onclick="onClick_${identifier}(event, '${result_tab_id}')" class='tablink-${identifier} active'>
                Result
            </a>
        </div>
        <div class="tablink-border-${identifier} log-tab-header">
            <a onclick="onClick_${identifier}(event, '${logs_tab_id}')" class='tablink-${identifier}'>Logs</a>
        </div>
    </div>
</div>
"""

result_html = """<div id="${result_tab_id}" class="tab-${identifier}" style="background: #F4F3F6;
    border-color: #CFCDD6; border-width: 0.5px; border-style: solid; padding: 24px; gap: 8px; margin-top: 24px">
    <div style="font-size: 12px; font-weight: 400; font: DejaVu Sans Mono, sans-serif; line-height: 16.8px">
        ${result}
    </div>
</div>
"""

job_repr_template = f"""
<!-- Start job_repr_template -->
<div class="syft-widget">

<!-- Start JS_DOWNLOAD_FONTS -->
{JS_DOWNLOAD_FONTS}
<!-- End JS_DOWNLOAD_FONTS -->

<!-- Start CSS_CODE -->
{CSS_CODE}
<!-- End CSS_CODE -->

<!-- Start header_line_html -->
{header_line_html}
<!-- End header_line_html -->

<!-- Start attrs_html -->
{attrs_html}
<!-- End attrs_html -->

<!-- Start tabs_html -->
{tabs_html}
<!-- End tabs_html -->

<!-- Start result_html -->
{result_html}
<!-- End result_html -->

<!-- Start logs_html -->
{logs_html}
<!-- End logs_html -->

<!-- Start onclick_html -->
{onclick_html}
<!-- End onclick_html -->

<div style='height: 16px;'></div>
</div>
<!-- End job_repr_template -->
"""



# stdlib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import smtplib

# third party
from pydantic import BaseModel
from result import Err
from result import Ok
from result import Result


class SMTPClient(BaseModel):
    SOCKET_TIMEOUT: int = 5  # seconds

    def __init__(
        self,
        server: str,
        port: int,
        username: str,
        password: str,
    ) -> None:
        if not (username and password):
            raise ValueError("Both username and password must be provided")

        self.username = username
        self.password = password
        self.server = server
        self.port = port

    def send(self, sender: str, receiver: list[str], subject: str, body: str) -> None:
        if not (subject and body and receiver):
            raise ValueError("Subject, body, and recipient email(s) are required")

        msg = MIMEMultipart("alternative")
        msg["From"] = sender
        msg["To"] = ", ".join(receiver)
        msg["Subject"] = subject
        msg.attach(MIMEText(body, "html"))

        with smtplib.SMTP(
            self.server, self.port, timeout=self.SOCKET_TIMEOUT
        ) as server:
            server.ehlo()
            if server.has_extn("STARTTLS"):
                server.starttls()
                server.ehlo()
            server.login(self.username, self.password)
            text = msg.as_string()
            server.sendmail(sender, ", ".join(receiver), text)
        # TODO: Add error handling

    @classmethod
    def check_credentials(
        cls, server: str, port: int, username: str, password: str
    ) -> Result[Ok, Err]:
        """Check if the credentials are valid.

        Returns:
            bool: True if the credentials are valid, False otherwise.
        """
        try:
            with smtplib.SMTP(server, port, timeout=cls.SOCKET_TIMEOUT) as smtp_server:
                smtp_server.ehlo()
                if smtp_server.has_extn("STARTTLS"):
                    smtp_server.starttls()
                    smtp_server.ehlo()
                smtp_server.login(username, password)
                return Ok("Credentials are valid.")
        except Exception as e:
            return Err(e)



# stdlib
from typing import Any
from typing import final

# third party
from pydantic import BaseModel
from pydantic._internal._model_construction import ModelMetaclass

# relative
from ..serde.serializable import serializable


class EmptyType(type):
    def __repr__(self) -> str:
        return self.__name__

    def __bool__(self) -> bool:
        return False


@serializable(canonical_name="Empty", version=1)
@final
class Empty(metaclass=EmptyType):
    pass


class PartialModelMetaclass(ModelMetaclass):
    def __new__(
        mcs,
        cls_name: str,
        bases: tuple[type[Any], ...],
        namespace: dict[str, Any],
        *args: Any,
        **kwargs: Any,
    ) -> type:
        cls = super().__new__(mcs, cls_name, bases, namespace, *args, **kwargs)

        if issubclass(cls, BaseModel):
            for field_info in cls.model_fields.values():
                if field_info.annotation is not None and field_info.is_required():
                    field_info.annotation = field_info.annotation | EmptyType
                    field_info.default = Empty

            cls.model_rebuild(force=True)

        return cls



# stdlib
from collections.abc import Callable
from typing import Any

# third party
from typing_extensions import Self

# relative
from ...serde.serializable import serializable
from ...store.document_store import PartitionKey
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.transforms import TransformContext
from ...types.transforms import add_server_uid_for_key
from ...types.transforms import generate_id
from ...types.transforms import transform
from ...types.uid import UID
from ...util.markdown import as_markdown_python_code
from ..response import SyftError

NamePartitionKey = PartitionKey(key="name", type_=str)


@serializable()
class DataSubject(SyftObject):
    # version
    __canonical_name__ = "DataSubject"
    __version__ = SYFT_OBJECT_VERSION_1

    server_uid: UID
    name: str
    description: str | None = None
    aliases: list[str] = []

    @property
    def members(self) -> list:
        # relative
        from ...client.api import APIRegistry

        api = APIRegistry.api_for(self.server_uid, self.syft_client_verify_key)
        if api is None:
            return SyftError(message=f"You must login to {self.server_uid}")
        members = api.services.data_subject.members_for(self.name)
        return members

    __attr_searchable__ = ["name", "description"]
    __repr_attrs__ = ["name", "description"]
    __attr_unique__ = ["name"]

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        return hash(self) == hash(other)

    def __repr_syft_nested__(self) -> str:
        return f"DataSubject({self.name})"

    def __repr__(self) -> str:
        return f"<DataSubject: {self.name}>"

    def _repr_markdown_(self, wrap_as_python: bool = True, indent: int = 0) -> str:
        _repr_str = f"DataSubject: {self.name}\n"
        _repr_str += f"Description: {self.description}\n"
        _repr_str += f"Aliases: {self.aliases}\n"
        _repr_str += f"Members: {len(self.members)}\n"
        return as_markdown_python_code(_repr_str)


@serializable()
class DataSubjectCreate(SyftObject):
    # version
    __canonical_name__ = "DataSubjectCreate"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID | None = None  # type: ignore[assignment]
    name: str
    description: str | None = None
    aliases: list[str] | None = []
    members: dict[str, "DataSubjectCreate"] = {}

    __attr_searchable__ = ["name", "description"]
    __attr_unique__ = ["name"]
    __repr_attrs__ = ["name", "member_count"]

    @property
    def member_count(self) -> int:
        return len(self.members)

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        return hash(self) == hash(other)

    def __repr_syft_nested__(self) -> str:
        return f"DataSubject({self.name})"

    def _create_member_relationship(
        self, data_subject: Self, _relationship_set: set
    ) -> None:
        for member in data_subject.members.values():
            _relationship_set.add((data_subject, member))
            self._create_member_relationship(member, _relationship_set)

    def add_member(self, data_subject: Self) -> None:
        self.members[data_subject.name] = data_subject

    @property
    def member_relationships(self) -> set[tuple[str, str]]:
        relationships: set = set()
        self._create_member_relationship(self, relationships)
        return relationships

    def __repr__(self) -> str:
        return f"<DataSubject: {self.name}>"

    def _repr_markdown_(self, wrap_as_python: bool = True, indent: int = 0) -> str:
        _repr_str = f"DataSubject: {self.name}\n"
        _repr_str += f"Description: {self.description}\n"
        _repr_str += f"Aliases: {self.aliases}\n"
        _repr_str += f"Members: {len(self.members)}\n"
        return as_markdown_python_code(_repr_str)


def remove_members_list(context: TransformContext) -> TransformContext:
    if context.output is not None:
        context.output.pop("members", [])
    return context


@transform(DataSubjectCreate, DataSubject)
def create_data_subject_to_data_subject() -> list[Callable]:
    return [generate_id, remove_members_list, add_server_uid_for_key("server_uid")]



# stdlib
from collections import OrderedDict
from collections import deque
from collections.abc import Callable
from collections.abc import Collection
from collections.abc import Iterable
from collections.abc import KeysView
from collections.abc import Mapping
from types import MappingProxyType
from typing import Generic
from typing import SupportsIndex
from typing import TypeVar
from typing import overload

# third party
from typing_extensions import Self

_T = TypeVar("_T")
_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


# To correctly implement the creation of a DictTuple instance with
# DictTuple(key_value_pairs: Iterable[tuple[_KT, _VT]]),
# implementing just __init__ and __new__ is not enough.
#
# We need to extract the keys and values from keys_value_pairs
# and pass the values to __new__ (needs to call tuple.__new__(values) to create the tuple)
# and the keys to __init__ to create the mapping _KT -> int,
# but we can only iterate over key_value_pairs once since keys_value_pairs,
# being just an Iterable, might be ephemeral.
#
# Implementing just __new__ and __init__ for DictTuple is not enough since when
# calling DictTuple(key_value_pairs), __new__(keys_and_values) and __init__(key_value_pairs)
# are called in 2 separate function calls. If keys_and_values are ephemeral, like a generator,
# by the time it's passed to __init__() it's already been exhausted and there is no way to
# extract the keys out to create the mapping.
#
# Thus it is necessary to override __call__ of the metaclass
# to customize the way __new__ and __init__ work together, by iterating over key_value_pairs
# once to extract both keys and values, then passing keys to __new__, values to __init__
# within the same function call.
class _Meta(type):
    @overload
    def __call__(cls: type[_T]) -> _T: ...

    @overload
    def __call__(cls: type[_T], __value: Iterable[tuple[_KT, _VT]]) -> _T: ...

    @overload
    def __call__(cls: type[_T], __value: Mapping[_KT, _VT]) -> _T: ...

    @overload
    def __call__(
        cls: type[_T], __value: Iterable[_VT], __key: Collection[_KT]
    ) -> _T: ...

    @overload
    def __call__(
        cls: type[_T], __value: Iterable[_VT], __key: Callable[[_VT], _KT]
    ) -> _T: ...

    def __call__(
        cls: type[_T],
        __value: Iterable | None = None,
        __key: Callable | Collection | None = None,
        /,
    ) -> _T:
        # DictTuple()
        if __value is None and __key is None:
            obj = cls.__new__(cls)
            obj.__init__()
            return obj

        # DictTuple(DictTuple(...))
        elif type(__value) is cls:
            return __value

        # DictTuple({"x": 123, "y": 456})
        elif isinstance(__value, Mapping) and __key is None:
            obj = cls.__new__(cls, __value.values())
            obj.__init__(__value.keys())

            return obj

        # DictTuple(EnhancedDictTuple(...))
        # EnhancedDictTuple(DictTuple(...))
        # where EnhancedDictTuple subclasses DictTuple
        elif hasattr(__value, "items") and callable(__value.items):
            return cls.__call__(__value.items())

        # DictTuple([("x", 123), ("y", 456)])
        elif isinstance(__value, Iterable) and __key is None:
            keys = OrderedDict()
            values = deque()

            for i, (k, v) in enumerate(__value):
                keys[k] = i
                values.append(v)

            obj = cls.__new__(cls, values)
            obj.__init__(keys)

            return obj

        # DictTuple([123, 456], ["x", "y"])
        elif isinstance(__value, Iterable) and isinstance(__key, Iterable):
            keys = OrderedDict((k, i) for i, k in enumerate(__key))

            obj = cls.__new__(cls, __value)
            obj.__init__(keys)

            return obj

        # DictTuple(["abc", "xyz"], lambda x: x[0])
        # equivalent to DictTuple({"a": "abc", "x": "xyz"})
        elif isinstance(__value, Iterable) and isinstance(__key, Callable):
            obj = cls.__new__(cls, __value)
            obj.__init__(__key)

            return obj

        raise NotImplementedError


class DictTuple(tuple[_VT, ...], Generic[_KT, _VT], metaclass=_Meta):
    """
    OVERVIEW

        tuple with support for dict-like __getitem__(key)

            dict_tuple = DictTuple({"x": 1, "y": 2})

            dict_tuple["x"] == 1

            dict_tuple["y"] == 2

            dict_tuple[0] == 1

            dict_tuple[1] == 2

        everything else, e.g. __contains__, __iter__, behaves similarly to a tuple


    CREATION

        DictTuple(iterable) -> DictTuple([("x", 1), ("y", 2)])

        DictTuple(mapping) -> DictTuple({"x": 1, "y": 2})

        DictTuple(values, keys) -> DictTuple([1, 2], ["x", "y"])


    IMPLEMENTATION DETAILS

        DictTuple[_KT, _VT] is essentially a tuple[_VT, ...] that maintains an immutable Mapping[_KT, int]
        from the key to the tuple index internally.

        For example DictTuple({"x": 12, "y": 34}) is just a tuple (12, 34) with a {"x": 0, "y": 1} mapping.

        types.MappingProxyType is used for the mapping for immutability.
    """

    __mapping: MappingProxyType[_KT, int]

    # These overloads are copied from _Meta.__call__ just for IDE hints
    @overload
    def __init__(self) -> None: ...

    @overload
    def __init__(self, __value: Iterable[tuple[_KT, _VT]]) -> None: ...

    @overload
    def __init__(self, __value: Mapping[_KT, _VT]) -> None: ...

    @overload
    def __init__(self, __value: Iterable[_VT], __key: Collection[_KT]) -> None: ...

    @overload
    def __init__(self, __value: Iterable[_VT], __key: Callable[[_VT], _KT]) -> None: ...

    def __init__(self, __value=None, /):
        if isinstance(__value, MappingProxyType):
            self.__mapping = __value
        elif isinstance(__value, Mapping):
            self.__mapping = MappingProxyType(__value)
        elif isinstance(__value, Iterable):
            self.__mapping = MappingProxyType(
                OrderedDict((k, i) for i, k in enumerate(__value))
            )
        elif isinstance(__value, Callable):
            self.__mapping = MappingProxyType(
                OrderedDict((__value(v), i) for i, v in enumerate(self))
            )

        super().__init__()

        if len(self.__mapping) != len(self):
            raise ValueError("`__keys` and `__values` do not have the same length")

        if any(isinstance(k, SupportsIndex) for k in self.__mapping.keys()):
            raise ValueError(
                "values of `__keys` should not have type `int`, "
                "or implement `__index__()`"
            )

    @overload
    def __getitem__(self, __key: _KT) -> _VT: ...

    @overload
    def __getitem__(self, __key: slice) -> Self: ...

    @overload
    def __getitem__(self, __key: SupportsIndex) -> _VT: ...

    def __getitem__(self, __key, /):
        if isinstance(__key, slice):
            return self.__class__(
                super().__getitem__(__key),
                list(self.__mapping.keys()).__getitem__(__key),
            )

        if isinstance(__key, SupportsIndex):
            return super().__getitem__(__key)

        return super().__getitem__(self.__mapping[__key])

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}{super().__repr__()}"

    def keys(self) -> KeysView[_KT]:
        return self.__mapping.keys()

    def items(self) -> Iterable[tuple[_KT, _VT]]:
        return zip(self.__mapping.keys(), self)



# stdlib
from typing import Any
from typing import ClassVar

# relative
from ...serde.serializable import serializable
from ...service.context import AuthedServiceContext
from ...store.document_store import BaseStash
from ...types.uid import UID
from ..response import SyftError
from ..response import SyftSuccess
from ..worker.worker_stash import WorkerStash


@serializable(canonical_name="QueueClientConfig", version=1)
class QueueClientConfig:
    pass


@serializable(canonical_name="AbstractMessageHandler", version=1)
class AbstractMessageHandler:
    queue_name: ClassVar[str]

    @staticmethod
    def handle_message(message: bytes, syft_worker_id: UID) -> None:
        raise NotImplementedError


@serializable(
    attrs=["message_handler", "queue_name", "address"],
    canonical_name="QueueConsumer",
    version=1,
)
class QueueConsumer:
    message_handler: AbstractMessageHandler
    queue_name: str
    address: str

    def receive(self) -> None:
        raise NotImplementedError

    def run(self) -> None:
        raise NotImplementedError

    def close(self) -> None:
        raise NotImplementedError


@serializable(canonical_name="QueueProducer", version=1)
class QueueProducer:
    queue_name: str

    @property
    def address(self) -> str:
        raise NotImplementedError

    def send(
        self,
        worker: bytes,
        message: Any,
    ) -> None:
        raise NotImplementedError

    def close(self) -> None:
        raise NotImplementedError


@serializable(canonical_name="QueueClient", version=1)
class QueueClient:
    def __init__(self, config: QueueClientConfig) -> None:
        raise NotImplementedError


@serializable(canonical_name="QueueConfig", version=1)
class QueueConfig:
    """Base Queue configuration"""

    client_type: type[QueueClient]
    client_config: QueueClientConfig


@serializable(canonical_name="BaseQueueManager", version=1)
class BaseQueueManager:
    config: QueueConfig

    def __init__(self, config: QueueConfig):
        self.config = config
        self.post_init()

    def post_init(self) -> None:
        pass

    def close(self) -> SyftError | SyftSuccess:
        raise NotImplementedError

    def create_consumer(
        self,
        message_handler: type[AbstractMessageHandler],
        service_name: str,
        worker_stash: WorkerStash | None = None,
        address: str | None = None,
        syft_worker_id: UID | None = None,
    ) -> QueueConsumer:
        raise NotImplementedError

    def create_producer(
        self,
        queue_name: str,
        queue_stash: type[BaseStash],
        context: AuthedServiceContext,
        worker_stash: WorkerStash,
    ) -> QueueProducer:
        raise NotImplementedError

    def send(self, message: bytes, queue_name: str) -> SyftSuccess | SyftError:
        raise NotImplementedError

    @property
    def publisher(self) -> QueueProducer:
        raise NotImplementedError
